"""Type supervisor.container.gt, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import validator

from gwatn import property_format
from gwatn.enums import SupervisorContainerStatus
from gwatn.property_format import predicate_validator


class SupervisorContainerStatus000SchemaEnum:
    enum_name: str = "supervisor.container.status.000"
    symbols: List[str] = [
        "00000000",
        "f48cff43",
        "17c5cc54",
        "ec342324",
        "cfde1b40",
        "4e28b6ae",
        "da2dafe0",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class SupervisorContainerStatus000(StrEnum):
    Unknown = auto()
    Authorized = auto()
    Launching = auto()
    Provisioning = auto()
    Running = auto()
    Stopped = auto()
    Deleted = auto()

    @classmethod
    def default(cls) -> "SupervisorContainerStatus000":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class SupervisorContainerStatusMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> SupervisorContainerStatus:
        if not SupervisorContainerStatus000SchemaEnum.is_symbol(symbol):
            raise SchemaError(
                f"{symbol} must belong to SupervisorContainerStatus000 symbols"
            )
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(
            versioned_enum,
            SupervisorContainerStatus,
            SupervisorContainerStatus.default(),
        )

    @classmethod
    def local_to_type(
        cls, supervisor_container_status: SupervisorContainerStatus
    ) -> str:
        if not isinstance(supervisor_container_status, SupervisorContainerStatus):
            raise SchemaError(
                f"{supervisor_container_status} must be of type {SupervisorContainerStatus}"
            )
        versioned_enum = as_enum(
            supervisor_container_status,
            SupervisorContainerStatus000,
            SupervisorContainerStatus000.default(),
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, SupervisorContainerStatus000] = {
        "00000000": SupervisorContainerStatus000.Unknown,
        "f48cff43": SupervisorContainerStatus000.Authorized,
        "17c5cc54": SupervisorContainerStatus000.Launching,
        "ec342324": SupervisorContainerStatus000.Provisioning,
        "cfde1b40": SupervisorContainerStatus000.Running,
        "4e28b6ae": SupervisorContainerStatus000.Stopped,
        "da2dafe0": SupervisorContainerStatus000.Deleted,
    }

    versioned_enum_to_type_dict: Dict[SupervisorContainerStatus000, str] = {
        SupervisorContainerStatus000.Unknown: "00000000",
        SupervisorContainerStatus000.Authorized: "f48cff43",
        SupervisorContainerStatus000.Launching: "17c5cc54",
        SupervisorContainerStatus000.Provisioning: "ec342324",
        SupervisorContainerStatus000.Running: "cfde1b40",
        SupervisorContainerStatus000.Stopped: "4e28b6ae",
        SupervisorContainerStatus000.Deleted: "da2dafe0",
    }


class SupervisorContainerGt(BaseModel):
    SupervisorContainerId: str  #
    Status: SupervisorContainerStatus  #
    WorldInstanceAlias: str  #
    SupervisorGNodeInstanceId: str  #
    SupervisorGNodeAlias: str  #
    TypeName: Literal["supervisor.container.gt"] = "supervisor.container.gt"
    Version: str = "000"

    _validator_supervisor_container_id = predicate_validator(
        "SupervisorContainerId", property_format.is_uuid_canonical_textual
    )

    @validator("Status")
    def _validator_status(
        cls, v: SupervisorContainerStatus
    ) -> SupervisorContainerStatus:
        return as_enum(v, SupervisorContainerStatus, SupervisorContainerStatus.Unknown)

    _validator_world_instance_alias = predicate_validator(
        "WorldInstanceAlias", property_format.is_world_instance_alias_format
    )

    _validator_supervisor_g_node_instance_id = predicate_validator(
        "SupervisorGNodeInstanceId", property_format.is_uuid_canonical_textual
    )

    _validator_supervisor_g_node_alias = predicate_validator(
        "SupervisorGNodeAlias", property_format.is_lrd_alias_format
    )

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["Status"]
        Status = as_enum(
            self.Status, SupervisorContainerStatus, SupervisorContainerStatus.default()
        )
        d["StatusGtEnumSymbol"] = SupervisorContainerStatusMap.local_to_type(Status)
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class SupervisorContainerGt_Maker:
    type_name = "supervisor.container.gt"
    version = "000"

    def __init__(
        self,
        supervisor_container_id: str,
        status: SupervisorContainerStatus,
        world_instance_alias: str,
        supervisor_g_node_instance_id: str,
        supervisor_g_node_alias: str,
    ):
        self.tuple = SupervisorContainerGt(
            SupervisorContainerId=supervisor_container_id,
            Status=status,
            WorldInstanceAlias=world_instance_alias,
            SupervisorGNodeInstanceId=supervisor_g_node_instance_id,
            SupervisorGNodeAlias=supervisor_g_node_alias,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: SupervisorContainerGt) -> str:
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> SupervisorContainerGt:
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> SupervisorContainerGt:
        d2 = dict(d)
        if "SupervisorContainerId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SupervisorContainerId")
        if "StatusGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StatusGtEnumSymbol")
        if d2["StatusGtEnumSymbol"] in SupervisorContainerStatus000SchemaEnum.symbols:
            d2["Status"] = SupervisorContainerStatusMap.type_to_local(
                d2["StatusGtEnumSymbol"]
            )
        else:
            d2["Status"] = SupervisorContainerStatus.default()
        if "WorldInstanceAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing WorldInstanceAlias")
        if "SupervisorGNodeInstanceId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SupervisorGNodeInstanceId")
        if "SupervisorGNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SupervisorGNodeAlias")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return SupervisorContainerGt(
            SupervisorContainerId=d2["SupervisorContainerId"],
            Status=d2["Status"],
            WorldInstanceAlias=d2["WorldInstanceAlias"],
            SupervisorGNodeInstanceId=d2["SupervisorGNodeInstanceId"],
            SupervisorGNodeAlias=d2["SupervisorGNodeAlias"],
            TypeName=d2["TypeName"],
            Version="000",
        )



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class Unit(StrEnum):
    Unknown = auto()
    Unitless = auto()
    W = auto()
    Celcius = auto()
    Fahrenheit = auto()
    Gpm = auto()
    Wh = auto()
    AmpsRms = auto()
    VoltsRms = auto()

    @classmethod
    def default(cls) -> "Unit":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



""" List of all the schema types """

from gwatn.schemata.accepted_bid import AcceptedBid
from gwatn.schemata.accepted_bid import AcceptedBid_Maker
from gwatn.schemata.atn_bid import AtnBid
from gwatn.schemata.atn_bid import AtnBid_Maker
from gwatn.schemata.atn_params_heatpumpwithbooststore import (
    AtnParamsHeatpumpwithbooststore,
)
from gwatn.schemata.atn_params_heatpumpwithbooststore import (
    AtnParamsHeatpumpwithbooststore_Maker,
)
from gwatn.schemata.atn_params_report_heatpumpwithbooststore import (
    AtnParamsReportHeatpumpwithbooststore,
)
from gwatn.schemata.atn_params_report_heatpumpwithbooststore import (
    AtnParamsReportHeatpumpwithbooststore_Maker,
)
from gwatn.schemata.atn_state_heatpumpwithbooststore import (
    AtnStateHeatpumpwithbooststore,
)
from gwatn.schemata.atn_state_heatpumpwithbooststore import (
    AtnStateHeatpumpwithbooststore_Maker,
)
from gwatn.schemata.bid_ack import BidAck
from gwatn.schemata.bid_ack import BidAck_Maker
from gwatn.schemata.flo_params_heatpumpwithbooststore import (
    FloParamsHeatpumpwithbooststore,
)
from gwatn.schemata.flo_params_heatpumpwithbooststore import (
    FloParamsHeatpumpwithbooststore_Maker,
)
from gwatn.schemata.g_node_gt import GNodeGt
from gwatn.schemata.g_node_gt import GNodeGt_Maker
from gwatn.schemata.g_node_instance_gt import GNodeInstanceGt
from gwatn.schemata.g_node_instance_gt import GNodeInstanceGt_Maker
from gwatn.schemata.heartbeat_a import HeartbeatA
from gwatn.schemata.heartbeat_a import HeartbeatA_Maker
from gwatn.schemata.latest_price import LatestPrice
from gwatn.schemata.latest_price import LatestPrice_Maker
from gwatn.schemata.market_price import MarketPrice
from gwatn.schemata.market_price import MarketPrice_Maker
from gwatn.schemata.market_slot import MarketSlot
from gwatn.schemata.market_slot import MarketSlot_Maker
from gwatn.schemata.market_type_gt import MarketTypeGt
from gwatn.schemata.market_type_gt import MarketTypeGt_Maker
from gwatn.schemata.price_quantity import PriceQuantity
from gwatn.schemata.price_quantity import PriceQuantity_Maker
from gwatn.schemata.price_quantity_unitless import PriceQuantityUnitless
from gwatn.schemata.price_quantity_unitless import PriceQuantityUnitless_Maker
from gwatn.schemata.ready import Ready
from gwatn.schemata.ready import Ready_Maker
from gwatn.schemata.sim_timestep import SimTimestep
from gwatn.schemata.sim_timestep import SimTimestep_Maker
from gwatn.schemata.super_starter import SuperStarter
from gwatn.schemata.super_starter import SuperStarter_Maker
from gwatn.schemata.supervisor_container_gt import SupervisorContainerGt
from gwatn.schemata.supervisor_container_gt import SupervisorContainerGt_Maker
from gwatn.schemata.tadeed_specs_hack import TadeedSpecsHack
from gwatn.schemata.tadeed_specs_hack import TadeedSpecsHack_Maker
from gwatn.schemata.terminalasset_certify_hack import TerminalassetCertifyHack
from gwatn.schemata.terminalasset_certify_hack import TerminalassetCertifyHack_Maker


__all__ = [
    "AcceptedBid",
    "AcceptedBid_Maker",
    "AtnBid",
    "AtnBid_Maker",
    "AtnParamsHeatpumpwithbooststore",
    "AtnParamsHeatpumpwithbooststore_Maker",
    "AtnParamsReportHeatpumpwithbooststore",
    "AtnParamsReportHeatpumpwithbooststore_Maker",
    "AtnStateHeatpumpwithbooststore",
    "AtnStateHeatpumpwithbooststore_Maker",
    "BidAck",
    "BidAck_Maker",
    "FloParamsHeatpumpwithbooststore",
    "FloParamsHeatpumpwithbooststore_Maker",
    "GNodeGt",
    "GNodeGt_Maker",
    "GNodeInstanceGt",
    "GNodeInstanceGt_Maker",
    "HeartbeatA",
    "HeartbeatA_Maker",
    "LatestPrice",
    "LatestPrice_Maker",
    "MarketPrice",
    "MarketPrice_Maker",
    "MarketSlot",
    "MarketSlot_Maker",
    "MarketTypeGt",
    "MarketTypeGt_Maker",
    "PriceQuantity",
    "PriceQuantity_Maker",
    "PriceQuantityUnitless",
    "PriceQuantityUnitless_Maker",
    "Ready",
    "Ready_Maker",
    "SimTimestep",
    "SimTimestep_Maker",
    "SuperStarter",
    "SuperStarter_Maker",
    "SupervisorContainerGt",
    "SupervisorContainerGt_Maker",
    "TadeedSpecsHack",
    "TadeedSpecsHack_Maker",
    "TerminalassetCertifyHack",
    "TerminalassetCertifyHack_Maker",
]



"""Type accepted.bid, version 000"""
import json
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import validator

from gwatn import property_format
from gwatn.property_format import predicate_validator
from gwatn.schemata.price_quantity_unitless import PriceQuantityUnitless
from gwatn.schemata.price_quantity_unitless import PriceQuantityUnitless_Maker


class AcceptedBid(BaseModel):
    MarketSlotName: str  #
    BidderAlias: str  #
    BidList: List[PriceQuantityUnitless]
    #
    ReceivedTimeUnixNs: int  #
    TypeName: Literal["accepted.bid"] = "accepted.bid"
    Version: str = "000"

    _validator_market_slot_name = predicate_validator(
        "MarketSlotName", property_format.is_market_slot_name_lrd_format
    )

    _validator_bidder_alias = predicate_validator(
        "BidderAlias", property_format.is_lrd_alias_format
    )

    @validator("BidList")
    def _validator_bid_list(cls, v: List) -> List:
        for elt in v:
            if not isinstance(elt, PriceQuantityUnitless):
                raise ValueError(
                    f"elt {elt} of BidList must have type PriceQuantityUnitless."
                )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()

        # Recursively call as_dict() for the SubTypes
        bid_list = []
        for elt in self.BidList:
            bid_list.append(elt.as_dict())
        d["BidList"] = bid_list
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class AcceptedBid_Maker:
    type_name = "accepted.bid"
    version = "000"

    def __init__(
        self,
        market_slot_name: str,
        bidder_alias: str,
        bid_list: List[PriceQuantityUnitless],
        received_time_unix_ns: int,
    ):
        self.tuple = AcceptedBid(
            MarketSlotName=market_slot_name,
            BidderAlias=bidder_alias,
            BidList=bid_list,
            ReceivedTimeUnixNs=received_time_unix_ns,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: AcceptedBid) -> str:
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> AcceptedBid:
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> AcceptedBid:
        d2 = dict(d)
        if "MarketSlotName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MarketSlotName")
        if "BidderAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing BidderAlias")
        if "BidList" not in d2.keys():
            raise SchemaError(f"dict {d2} missing BidList")
        bid_list = []
        if not isinstance(d2["BidList"], List):
            raise SchemaError("BidList must be a List!")
        for elt in d2["BidList"]:
            if not isinstance(elt, dict):
                raise SchemaError(
                    f"elt {elt} of BidList must be "
                    "PriceQuantityUnitless but not even a dict!"
                )
            bid_list.append(PriceQuantityUnitless_Maker.dict_to_tuple(elt))
        d2["BidList"] = bid_list
        if "ReceivedTimeUnixNs" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ReceivedTimeUnixNs")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return AcceptedBid(
            MarketSlotName=d2["MarketSlotName"],
            BidderAlias=d2["BidderAlias"],
            BidList=d2["BidList"],
            ReceivedTimeUnixNs=d2["ReceivedTimeUnixNs"],
            TypeName=d2["TypeName"],
            Version="000",
        )



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class GNodeRole(StrEnum):
    GNode = auto()
    TerminalAsset = auto()
    Scada = auto()
    PriceService = auto()
    WeatherService = auto()
    AtomicTNode = auto()
    MarketMaker = auto()
    AtomicMeteringNode = auto()
    ConductorTopologyNode = auto()
    InterconnectionComponent = auto()
    World = auto()
    TimeCoordinator = auto()
    Supervisor = auto()

    @classmethod
    def default(cls) -> "GNodeRole":
        return cls.GNode

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class ActorClass(StrEnum):
    NoActor = auto()
    PrimaryScada = auto()
    PrimaryMeter = auto()
    BooleanActuator = auto()
    SimpleSensor = auto()
    HomeAlone = auto()
    Atn = auto()
    MultipurposeSensor = auto()
    Thermostat = auto()

    @classmethod
    def default(cls) -> "ActorClass":
        return cls.NoActor

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class AtnSpaceheatStrategyName(StrEnum):
    NoActor = auto()
    SupervisorA = auto()
    HeatPumpWithBoostStore = auto()

    @classmethod
    def default(cls) -> "AtnSpaceheatStrategyName":
        return cls.NoActor

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



"""Type heartbeat.a, version 001"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import validator


class HeartbeatA(BaseModel):
    MyHex: str  #
    YourLastHex: str  #
    TypeName: Literal["heartbeat.a"] = "heartbeat.a"
    Version: str = "001"

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class HeartbeatA_Maker:
    type_name = "heartbeat.a"
    version = "001"

    def __init__(self, my_hex: str, your_last_hex: str):
        self.tuple = HeartbeatA(
            MyHex=my_hex,
            YourLastHex=your_last_hex,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: HeartbeatA) -> str:
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> HeartbeatA:
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> HeartbeatA:
        d2 = dict(d)
        if "MyHex" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MyHex")
        if "YourLastHex" not in d2.keys():
            raise SchemaError(f"dict {d2} missing YourLastHex")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return HeartbeatA(
            MyHex=d2["MyHex"],
            YourLastHex=d2["YourLastHex"],
            TypeName=d2["TypeName"],
            Version="001",
        )



import csv
import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple

import numpy as np
from gridworks.errors import SchemaError

import gwatn.property_format as property_format


class CsvWeatherForecast(NamedTuple):
    LocationAlias: str
    Comment: str
    StartYearUtc: int
    StartMonthUtc: int
    StartDayUtc: int
    StartHourUtc: int
    UniformSliceDurationHrs: float
    TimezoneString: str
    WeatherUid: str
    TempUnit: Optional[str]
    ForecastUnixTimeS: int
    IrradianceType: Optional[str] = None
    TempSourceAlias: Optional[str] = None
    TempMethodAlias: Optional[str] = None
    IrradianceSourceAlias: Optional[str] = None
    IrradianceMethodAlias: Optional[str] = None
    SkyClaritySourceAlias: Optional[str] = None
    SkyClarityMethodAlias: Optional[str] = None
    Temperatures: Optional[list] = None
    IrradianceWPerM2: Optional[List[int]] = None
    OpaqueCloudCoveragePercents: Optional[List[int]] = None
    TypeName: str = "csv.weather.forecast.sync"

    def asdict(self):
        d = self._asdict()
        if d["IrradianceType"] is None:
            del d["IrradianceType"]
        if d["IrradianceSourceAlias"] is None:
            del d["IrradianceSourceAlias"]
        if d["SkyClarityMethodAlias"] is None:
            del d["SkyClarityMethodAlias"]
        if d["TempSourceAlias"] is None:
            del d["TempSourceAlias"]
        if d["TempMethodAlias"] is None:
            del d["TempMethodAlias"]
        if d["IrradianceWPerM2"] is None:
            del d["IrradianceWPerM2"]
        if d["TempUnit"] is None:
            del d["TempUnit"]
        if d["SkyClaritySourceAlias"] is None:
            del d["SkyClaritySourceAlias"]
        if d["IrradianceMethodAlias"] is None:
            del d["IrradianceMethodAlias"]
        if d["OpaqueCloudCoveragePercents"] is None:
            del d["OpaqueCloudCoveragePercents"]
        if d["Temperatures"] is None:
            del d["Temperatures"]
        return d

    def is_valid(self) -> Tuple[bool, Optional[List[str]]]:
        is_valid, errors = self.passes_derived_validations()
        return is_valid, errors

    def passes_derived_validations(self) -> Tuple[bool, Optional[List[str]]]:
        is_valid = True
        errors = []
        if self.TypeName != "csv.weather.forecast.sync":
            is_valid = False
            errors.append(
                f"Payload requires TypeName of 'csv.weather.forecast.sync', not {self.TypeName}."
            )
        if not isinstance(self.LocationAlias, str):
            is_valid = False
            errors.append(f"LocationAlias must have type str.")
        if not isinstance(self.StartYearUtc, int):
            is_valid = False
            errors.append(f"StartYearUtc must have type int.")
        if not isinstance(self.StartMonthUtc, int):
            is_valid = False
            errors.append(f"StartMonthUtc must have type int.")
        if not isinstance(self.StartDayUtc, int):
            is_valid = False
            errors.append(f"StartDayUtc must have type int.")
        if not isinstance(self.StartHourUtc, int):
            is_valid = False
            errors.append(f"StartHourUtc must have type int.")
        try:
            datetime.datetime(
                year=self.StartYearUtc,
                month=self.StartMonthUtc,
                day=self.StartDayUtc,
                hour=self.StartHourUtc,
            )
        except ValueError as e:
            is_valid = False
            errors.append(e)
        if not isinstance(self.UniformSliceDurationHrs, float):
            is_valid = False
            errors.append(f"UniformSliceDurationHrs must have type float.")
        if not isinstance(self.TimezoneString, str):
            is_valid = False
            errors.append(f"TimezoneString {self.TimezoneString} must have type str.")
        if not isinstance(self.WeatherUid, str):
            is_valid = False
            errors.append(f"WeatherUid {self.WeatherUid} must have type str.")
        if not property_format.is_uuid_canonical_textual(self.WeatherUid):
            is_valid = False
            errors.append(
                f"WeatherUid {self.WeatherUid} must have format UuidCanonicalTextual."
            )

        return is_valid, errors


class CsvWeatherForecastSync_Maker:
    mp_alias = "csv.weather.forecast.sync"

    @classmethod
    def payload_is_valid(
        cls, payload_as_dict: Dict[str, Any]
    ) -> Tuple[bool, Optional[List[str]]]:
        try:
            payload_as_dict["UniformSliceDurationHrs"] = float(
                payload_as_dict["UniformSliceDurationHrs"]
            )
        except ValueError:
            pass  # This will get caught in is_valid() check below
        if "TempUnit" not in payload_as_dict.keys():
            payload_as_dict["TempUnit"] = None
        if "TempSourceAlias" not in payload_as_dict.keys():
            payload_as_dict["TempSourceAlias"] = None
        if "TempMethodAlias" not in payload_as_dict.keys():
            payload_as_dict["TempMethodAlias"] = None
        if "IrradianceType" not in payload_as_dict.keys():
            payload_as_dict["IrradianceType"] = None
        if "IrradianceSourceAlias" not in payload_as_dict.keys():
            payload_as_dict["IrradianceSourceAlias"] = None
        if "IrradianceMethodAlias" not in payload_as_dict.keys():
            payload_as_dict["IrradianceMethodAlias"] = None
        if "SkyClaritySourceAlias" not in payload_as_dict.keys():
            payload_as_dict["SkyClaritySourceAlias"] = None
        if "SkyClarityMethodAlias" not in payload_as_dict.keys():
            payload_as_dict["SkyClarityMethodAlias"] = None
        if "Temperatures" not in payload_as_dict.keys():
            payload_as_dict["Temperatures"] = None
        if "IrradianceWPerM2" not in payload_as_dict.keys():
            payload_as_dict["IrradianceWPerM2"] = None
        if "OpaqueCloudCoveragePercents" not in payload_as_dict.keys():
            payload_as_dict["OpaqueCloudCoveragePercents"] = None

        try:
            p = CsvWeatherForecast(
                TypeName=payload_as_dict["TypeName"],
                LocationAlias=payload_as_dict["LocationAlias"],
                Comment=payload_as_dict["Comment"],
                StartYearUtc=payload_as_dict["StartYearUtc"],
                StartMonthUtc=payload_as_dict["StartMonthUtc"],
                StartDayUtc=payload_as_dict["StartDayUtc"],
                StartHourUtc=payload_as_dict["StartHourUtc"],
                UniformSliceDurationHrs=payload_as_dict["UniformSliceDurationHrs"],
                TimezoneString=payload_as_dict["TimezoneString"],
                ForecastUnixTimeS=payload_as_dict["ForecastUnixTimeS"],
                WeatherUid=payload_as_dict["WeatherUid"],
                TempUnit=payload_as_dict["TempUnit"],
                IrradianceType=payload_as_dict["IrradianceType"],
                TempSourceAlias=payload_as_dict["TempSourceAlias"],
                TempMethodAlias=payload_as_dict["TempMethodAlias"],
                IrradianceSourceAlias=payload_as_dict["IrradianceSourceAlias"],
                IrradianceMethodAlias=payload_as_dict["IrradianceMethodAlias"],
                SkyClaritySourceAlias=payload_as_dict["SkyClaritySourceAlias"],
                SkyClarityMethodAlias=payload_as_dict["SkyClarityMethodAlias"],
                Temperatures=payload_as_dict["Temperatures"],
                IrradianceWPerM2=payload_as_dict["IrradianceWPerM2"],
                OpaqueCloudCoveragePercents=payload_as_dict[
                    "OpaqueCloudCoveragePercents"
                ],
            )
        except TypeError:
            errors = [TypeError]
            return False, errors
        return p.is_valid()

    def __init__(
        self,
        weather_csv,
    ):
        self.errors = []
        self.tuple: Optional[CsvWeatherForecast] = None
        first_data_row = 22
        temperatures = []
        irradiances = []
        sky_clarities = []
        uniform_slice_duration_hrs = np.array([])
        with open(weather_csv, newline="") as csvfile:
            reader = csv.reader(csvfile, delimiter=",")
            line_idx = 0
            for row in reader:
                if line_idx == 0:
                    property = row[0].replace("\ufeff", "")
                    mp_alias = row[1]
                    if property != "TypeName":
                        raise Exception(
                            f"Csv 0,0 must be TypeName, instead it is {property}"
                        )
                    if mp_alias != "csv.weather.forecast.sync":
                        raise Exception(
                            f"TypeName for {weather_csv} must be csv.weather.forecast.sync. Instead it was {mp_alias}"
                        )
                if line_idx == 1:
                    property = row[0]
                    location_alias = row[1].strip()
                    if property != "LocationAlias":
                        raise Exception(
                            f"Csv 1,0 must be LocationAlias, instead it is {property}"
                        )
                if line_idx == 2:
                    property = row[0]
                    if len(row) == 1:
                        temp_source_alias = None
                    else:
                        temp_source_alias = row[1].strip()
                        if temp_source_alias == "":
                            temp_source_alias = None
                    if property != "TempSourceAlias":
                        raise Exception(
                            f"Csv 2,0 must be TempSourceAlias, instead it is {property}"
                        )
                if line_idx == 3:
                    property = row[0]
                    temp_method_alias = row[1].strip()
                    if len(row) == 1:
                        temp_method_alias = None
                    else:
                        if temp_method_alias == "":
                            temp_method_alias = None
                    if property != "TempMethodAlias":
                        raise Exception(
                            f"Csv 3,0 must be TempMethodAlias, instead it is {property}"
                        )
                if line_idx == 4:
                    property = row[0]
                    if len(row) == 1:
                        sky_clarity_source_alias = None
                    else:
                        sky_clarity_source_alias = row[1].strip()
                        if sky_clarity_source_alias == "":
                            sky_clarity_source_alias = None
                    if property != "SkyClaritySourceAlias":
                        raise Exception(
                            f"Csv 4,0 must be SkyClaritySourceAlias, instead it is {property}"
                        )
                if line_idx == 5:
                    property = row[0]
                    if len(row) == 1:
                        sky_clarity_method_alias = None
                    else:
                        sky_clarity_method_alias = row[1].strip()
                        if sky_clarity_method_alias == "":
                            sky_clarity_method_alias = None
                    if property != "SkyClarityMethodAlias":
                        raise Exception(
                            f"Csv 5,0 must be SkyClarityMethodAlias, instead it is {property}"
                        )
                if line_idx == 6:
                    property = row[0]
                    if len(row) == 1:
                        irradiance_source_alias = None
                    else:
                        irradiance_source_alias = row[1].strip()
                        if irradiance_source_alias == "":
                            irradiance_source_alias = None
                    if property != "IrradianceSourceAlias":
                        raise Exception(
                            f"Csv 6,0 of {weather_csv} must be IrradianceSourceAlias, instead it is {property}"
                        )
                if line_idx == 7:
                    property = row[0]
                    if len(row) == 1:
                        irradiance_method_alias = None
                    else:
                        irradiance_method_alias = row[1].strip()
                        if irradiance_method_alias == "":
                            irradiance_method_alias = None
                    if property != "IrradianceMethodAlias":
                        raise Exception(
                            f"Csv 7,0 must be IrradianceMethodAlias, instead it is {property}"
                        )
                if line_idx == 8:
                    property = row[0]
                    comment = row[1].strip()
                    if property != "Comment":
                        raise Exception(
                            f"Csv 8,0 must be Comment, instead it is {property}"
                        )
                if line_idx == 9:
                    property = row[0]
                    start_year_utc = int(row[1])
                    if property != "StartYearUtc":
                        raise Exception(
                            f"Csv 9,0 must be StartYearUtc, instead it is {property}"
                        )
                if line_idx == 10:
                    property = row[0]
                    start_month_utc = int(row[1])
                    if property != "StartMonthUtc":
                        raise Exception(
                            f"Csv 10,0 must be StartMonthUtc, instead it is {property}"
                        )
                if line_idx == 11:
                    property = row[0]
                    start_day_utc = int(row[1])
                    if property != "StartDayUtc":
                        raise Exception(
                            f"Csv 11,0 must be StartDayUtc, instead it is {property}"
                        )
                if line_idx == 12:
                    property = row[0]
                    start_hour_utc = int(row[1])
                    if property != "StartHourUtc":
                        raise Exception(
                            f"Csv 12,0 must be StartHourUtc, instead it is {property}"
                        )
                if line_idx == 13:
                    property = row[0]
                    uniform_slice_duration_hrs = float(row[1])
                    if property != "UniformSliceDurationHrs":
                        raise Exception(
                            f"Csv 13,0 must be UniformSliceDurationHrs, instead it is {property}"
                        )
                if line_idx == 14:
                    property = row[0]
                    timezone_string = row[1].strip()
                    if property != "TimezoneString":
                        raise Exception(
                            f"Csv 14,0 must be TimezoneString, instead it is {property}"
                        )
                if line_idx == 15:
                    property = row[0]
                    if len(row) == 1:
                        temp_unit = None
                    else:
                        temp_unit = row[1].strip()
                        if temp_unit == "":
                            temp_unit = None
                    if property != "TempUnit":
                        raise Exception(
                            f"Csv 15,0 must be TempUnit, instead it is {property}"
                        )
                if line_idx == 16:
                    property = row[0]
                    if len(row) == 1:
                        irradiance_type = None
                    else:
                        irradiance_type = row[1].strip()
                        if irradiance_type == "":
                            irradiance_type = None
                    if property != "IrradianceType":
                        raise Exception(
                            f"Csv 16,0 must be IrradianceType, instead it is {property}"
                        )
                if line_idx == 17:
                    property = row[0]
                    include_temp_str = row[1].strip()
                    if property != "IncludeTemp":
                        raise Exception(
                            f"Csv 17,0 must be IncludeTemp, instead it is {property}"
                        )
                    if include_temp_str == "True":
                        include_temp = True
                    else:
                        include_temp = False
                if line_idx == 18:
                    property = row[0]
                    include_irradiance_str = row[1].strip()
                    if property != "IncludeIrradiance":
                        raise Exception(
                            f"Csv 18,0 must be IncludeIrradiance, instead it is {property}"
                        )
                    if include_irradiance_str == "True":
                        include_irradiance = True
                    else:
                        include_irradiance = False
                if line_idx == 19:
                    property = row[0]
                    include_sky_clarity_str = row[1].strip()
                    if property != "IncludeSkyClarity":
                        raise Exception(
                            f"Csv 19,0 must be IncludeSkyClarity, instead it is {property}"
                        )
                    if include_sky_clarity_str == "True":
                        include_sky_clarity = True
                    else:
                        include_sky_clarity = False
                if line_idx == 20:
                    property = row[0]
                    forecast_unix_time_s = int(row[1])
                    if property != "ForecastUnixTimeS":
                        raise Exception(
                            f"Csv 20,0 must be ForecastUnixTimeS, instead it is {property}"
                        )
                if line_idx == 21:
                    property = row[0]
                    price_uid = row[1].strip()
                    if property != "WeatherUid":
                        raise Exception(
                            f"Csv 21,0 must be WeatherUid, instead it is {property}"
                        )
                if line_idx == 22:
                    header0 = row[0].strip()
                    header1 = row[1].strip()
                    header2 = row[2].strip()
                    if header0 != "Temperatures":
                        raise Exception(
                            f"Csv 22,0 must be Temperatures, instead it is {header0}"
                        )
                    if header1 != "IrradianceWPerM2":
                        raise Exception(
                            f"Csv 22,1 must be IrradianceWPerM2, instead it is {header1}"
                        )
                    if header2 != "OpaqueCloudCoveragePercents":
                        raise Exception(
                            f"Csv 22,2 must be OpaqueCloudCoveragePercents, instead it is {header2}"
                        )
                elif line_idx >= first_data_row:
                    if include_temp:
                        try:
                            temperatures.append(int(row[0]))
                        except ValueError:
                            raise Exception(
                                f"File {weather_csv} Missing a temperature in row {line_idx+1}"
                            )
                    if include_irradiance:
                        try:
                            irradiances.append(int(row[1]))
                        except ValueError:
                            raise Exception(
                                f"Missing an irradiance in row {line_idx+1}"
                            )
                    if include_sky_clarity:
                        try:
                            sky_clarities.append(int(row[2]))
                        except ValueError:
                            raise Exception(f"Missing an occp in row {line_idx+1}")
                line_idx += 1
        if not include_temp:
            temperatures = None
            temp_unit = None
        if not include_irradiance:
            irradiances = None
            irradiance_type = None
        if not include_sky_clarity:
            sky_clarities = None

        p = CsvWeatherForecast(
            LocationAlias=location_alias,
            Comment=comment,
            StartYearUtc=start_year_utc,
            StartMonthUtc=start_month_utc,
            StartDayUtc=start_day_utc,
            StartHourUtc=start_hour_utc,
            UniformSliceDurationHrs=uniform_slice_duration_hrs,
            TimezoneString=timezone_string,
            ForecastUnixTimeS=forecast_unix_time_s,
            WeatherUid=price_uid,
            TempUnit=temp_unit,
            IrradianceType=irradiance_type,
            TempSourceAlias=temp_source_alias,
            TempMethodAlias=temp_method_alias,
            IrradianceSourceAlias=irradiance_source_alias,
            IrradianceMethodAlias=irradiance_method_alias,
            SkyClaritySourceAlias=sky_clarity_source_alias,
            SkyClarityMethodAlias=sky_clarity_method_alias,
            Temperatures=temperatures,
            IrradianceWPerM2=irradiances,
            OpaqueCloudCoveragePercents=sky_clarities,
        )

        is_valid, errors = p.is_valid()
        if is_valid is False:
            self.errors = errors
            print(errors)
            raise SchemaError(
                f"Failed to create payload due to these errors:{errors}. Input file is {weather_csv}"
            )

        self.tuple = p



""" List of all the types used"""
from typing import Dict
from typing import List
from typing import no_type_check

from gwatn.schemata import AcceptedBid_Maker
from gwatn.schemata import AtnBid_Maker
from gwatn.schemata import AtnParamsHeatpumpwithbooststore_Maker
from gwatn.schemata import AtnParamsReportHeatpumpwithbooststore_Maker
from gwatn.schemata import AtnStateHeatpumpwithbooststore_Maker
from gwatn.schemata import BidAck_Maker
from gwatn.schemata import FloParamsHeatpumpwithbooststore_Maker
from gwatn.schemata import GNodeGt_Maker
from gwatn.schemata import GNodeInstanceGt_Maker
from gwatn.schemata import HeartbeatA_Maker
from gwatn.schemata import LatestPrice_Maker
from gwatn.schemata import MarketPrice_Maker
from gwatn.schemata import MarketSlot_Maker
from gwatn.schemata import MarketTypeGt_Maker
from gwatn.schemata import PriceQuantity_Maker
from gwatn.schemata import PriceQuantityUnitless_Maker
from gwatn.schemata import Ready_Maker
from gwatn.schemata import SimTimestep_Maker
from gwatn.schemata import SuperStarter_Maker
from gwatn.schemata import SupervisorContainerGt_Maker
from gwatn.schemata import TadeedSpecsHack_Maker
from gwatn.schemata import TerminalassetCertifyHack_Maker


TypeMakerByName: Dict[str, HeartbeatA_Maker] = {}


@no_type_check
def type_makers() -> List[HeartbeatA_Maker]:
    return [
        AcceptedBid_Maker,
        AtnBid_Maker,
        AtnParamsHeatpumpwithbooststore_Maker,
        AtnParamsReportHeatpumpwithbooststore_Maker,
        AtnStateHeatpumpwithbooststore_Maker,
        BidAck_Maker,
        FloParamsHeatpumpwithbooststore_Maker,
        GNodeGt_Maker,
        GNodeInstanceGt_Maker,
        HeartbeatA_Maker,
        LatestPrice_Maker,
        MarketPrice_Maker,
        MarketSlot_Maker,
        MarketTypeGt_Maker,
        PriceQuantity_Maker,
        PriceQuantityUnitless_Maker,
        Ready_Maker,
        SimTimestep_Maker,
        SuperStarter_Maker,
        SupervisorContainerGt_Maker,
        TadeedSpecsHack_Maker,
        TerminalassetCertifyHack_Maker,
    ]


for maker in type_makers():
    TypeMakerByName[maker.type_name] = maker


def version_by_type_name() -> Dict[str, str]:
    """
    Returns:
        Dict[str, str]: Keys are TypeNames, values are versions
    """

    v: Dict[str, str] = {
        "accepted.bid": "000",
        "atn.bid": "000",
        "atn.params.heatpumpwithbooststore": "000",
        "atn.params.report.heatpumpwithbooststore": "000",
        "atn.state.heatpumpwithbooststore": "000",
        "bid.ack": "000",
        "flo.params.heatpumpwithbooststore": "000",
        "g.node.gt": "000",
        "g.node.instance.gt": "000",
        "heartbeat.a": "001",
        "latest.price": "000",
        "market.price": "000",
        "market.slot": "000",
        "market.type.gt": "000",
        "price.quantity": "000",
        "price.quantity.unitless": "000",
        "ready": "001",
        "sim.timestep": "000",
        "super.starter": "000",
        "supervisor.container.gt": "000",
        "tadeed.specs.hack": "000",
        "terminalasset.certify.hack": "000",
    }

    return v


def status_by_versioned_type_name() -> Dict[str, str]:
    """
    Returns:
        Dict[str, str]: Keys are versioned TypeNames, values are type status
    """

    v: Dict[str, str] = {
        "accepted.bid.000": "Pending",
        "atn.bid.000": "Pending",
        "atn.params.heatpumpwithbooststore.000": "Active",
        "atn.params.report.heatpumpwithbooststore.000": "Pending",
        "atn.state.heatpumpwithbooststore.000": "Pending",
        "bid.ack.000": "Pending",
        "flo.params.heatpumpwithbooststore.000": "Active",
        "g.node.gt.000": "Pending",
        "g.node.instance.gt.000": "Pending",
        "heartbeat.a.001": "Pending",
        "latest.price.000": "Pending",
        "market.price.000": "Pending",
        "market.slot.000": "Pending",
        "market.type.gt.000": "Pending",
        "price.quantity.000": "Pending",
        "price.quantity.unitless.000": "Pending",
        "ready.001": "Pending",
        "sim.timestep.000": "Pending",
        "super.starter.000": "Pending",
        "supervisor.container.gt.000": "Pending",
        "tadeed.specs.hack.000": "Pending",
        "terminalasset.certify.hack.000": "Pending",
    }

    return v



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class GniStatus(StrEnum):
    Unknown = auto()
    Pending = auto()
    Active = auto()
    Done = auto()

    @classmethod
    def default(cls) -> "GniStatus":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MarketTypeName(StrEnum):
    unknown = auto()
    rt5gate5 = auto()
    rt60gate5 = auto()
    da60 = auto()
    rt60gate30 = auto()
    rt15gate5 = auto()
    rt30gate5 = auto()
    rt60gate30b = auto()

    @classmethod
    def default(cls) -> "MarketTypeName":
        return cls.unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



"""Settings for an AtomicTNode, readable from environment and/or from env files."""

import pendulum
from gridworks.gw_config import GNodeSettings


DEFAULT_ENV_FILE = ".env"


class Settings(GNodeSettings):
    g_node_alias: str = "d1.isone.ver.keene.holly"
    g_node_role_value: str = "AtomicTNode"
    my_super_alias: str = "d1.isone.ver.keene.super1"

    market_maker_alias = "d1.isone.ver.keene"
    market_maker_algo_address = (
        "JMEUH2AXM6UGRJO2DBZXDOA2OMIWQFNQZ54LCVC4GQX6QDOX5Z6JRGMWFA"
    )
    mm_api_root = "http://localhost:7997"
    initial_time_unix_s = pendulum.datetime(
        year=2020, month=1, day=1, hour=4, minute=20
    ).int_timestamp

    class Config:
        env_prefix = "ATN_"
        env_nested_delimiter = "__"


class SupervisorSettings(GNodeSettings):
    g_node_alias: str = "d1.isone.ver.keene.super1"
    g_node_role_value: str = "Supervisor"

    class Config:
        env_prefix = "SUPER_"
        env_nested_delimiter = "__"



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MakeModel(StrEnum):
    FREEDOM__RELAY = auto()
    OMEGAEZO__FTB800FLO = auto()
    UNKNOWNMAKE__UNKNOWNMODEL = auto()
    GRIDWORKS__SIMBOOL30AMPRELAY = auto()
    MAGNELAB__SCT0300050 = auto()
    YMDC__SCT013100 = auto()
    G1__NCD_ADS1115__TEWA_NTC_10K_A = auto()
    G1__NCD_ADS1115__AMPH_NTC_10K_A = auto()
    GRIDWORKS__SIMPM1 = auto()
    GRIDWORKS__WATERTEMPHIGHPRECISION = auto()
    NCD__PR814SPST = auto()
    ADAFRUIT__642 = auto()
    SCHNEIDERELECTRIC__IEM3455 = auto()
    OPENENERGY__EMONPI = auto()
    EGAUGE__3010 = auto()
    RHEEM__XE50T10H45U0 = auto()

    @classmethod
    def default(cls) -> "MakeModel":
        return cls.UNKNOWNMAKE__UNKNOWNMODEL

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



"""Type g.node.instance.gt, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import validator

from gwatn import property_format
from gwatn.data_classes.g_node_instance import GNodeInstance
from gwatn.enums import GniStatus
from gwatn.enums import StrategyName
from gwatn.property_format import predicate_validator
from gwatn.schemata.g_node_gt import GNodeGt
from gwatn.schemata.g_node_gt import GNodeGt_Maker


class GniStatus000SchemaEnum:
    enum_name: str = "gni.status.000"
    symbols: List[str] = [
        "00000000",
        "7890ab0a",
        "69241259",
        "8222421f",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class GniStatus000(StrEnum):
    Unknown = auto()
    Pending = auto()
    Active = auto()
    Done = auto()

    @classmethod
    def default(cls) -> "GniStatus000":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class GniStatusMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> GniStatus:
        if not GniStatus000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to GniStatus000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, GniStatus, GniStatus.default())

    @classmethod
    def local_to_type(cls, gni_status: GniStatus) -> str:
        if not isinstance(gni_status, GniStatus):
            raise SchemaError(f"{gni_status} must be of type {GniStatus}")
        versioned_enum = as_enum(gni_status, GniStatus000, GniStatus000.default())
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, GniStatus000] = {
        "00000000": GniStatus000.Unknown,
        "7890ab0a": GniStatus000.Pending,
        "69241259": GniStatus000.Active,
        "8222421f": GniStatus000.Done,
    }

    versioned_enum_to_type_dict: Dict[GniStatus000, str] = {
        GniStatus000.Unknown: "00000000",
        GniStatus000.Pending: "7890ab0a",
        GniStatus000.Active: "69241259",
        GniStatus000.Done: "8222421f",
    }


class StrategyName000SchemaEnum:
    enum_name: str = "strategy.name.000"
    symbols: List[str] = [
        "00000000",
        "642c83d3",
        "4bb2cf7e",
        "f5961401",
        "73fbe6ab",
        "5e18a52e",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class StrategyName000(StrEnum):
    NoActor = auto()
    WorldA = auto()
    SupervisorA = auto()
    AtnHeatPumpWithBoostStore = auto()
    TcGlobalA = auto()
    MarketMakerA = auto()

    @classmethod
    def default(cls) -> "StrategyName000":
        return cls.NoActor

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class StrategyNameMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> StrategyName:
        if not StrategyName000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to StrategyName000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, StrategyName, StrategyName.default())

    @classmethod
    def local_to_type(cls, strategy_name: StrategyName) -> str:
        if not isinstance(strategy_name, StrategyName):
            raise SchemaError(f"{strategy_name} must be of type {StrategyName}")
        versioned_enum = as_enum(
            strategy_name, StrategyName000, StrategyName000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, StrategyName000] = {
        "00000000": StrategyName000.NoActor,
        "642c83d3": StrategyName000.WorldA,
        "4bb2cf7e": StrategyName000.SupervisorA,
        "f5961401": StrategyName000.AtnHeatPumpWithBoostStore,
        "73fbe6ab": StrategyName000.TcGlobalA,
        "5e18a52e": StrategyName000.MarketMakerA,
    }

    versioned_enum_to_type_dict: Dict[StrategyName000, str] = {
        StrategyName000.NoActor: "00000000",
        StrategyName000.WorldA: "642c83d3",
        StrategyName000.SupervisorA: "4bb2cf7e",
        StrategyName000.AtnHeatPumpWithBoostStore: "f5961401",
        StrategyName000.TcGlobalA: "73fbe6ab",
        StrategyName000.MarketMakerA: "5e18a52e",
    }


class GNodeInstanceGt(BaseModel):
    GNodeInstanceId: str  #
    GNode: GNodeGt  #
    Strategy: StrategyName  #
    Status: GniStatus  #
    SupervisorContainerId: str  #
    StartTimeUnixS: int  #
    EndTimeUnixS: int  #
    AlgoAddress: Optional[str] = None
    TypeName: Literal["g.node.instance.gt"] = "g.node.instance.gt"
    Version: str = "000"

    _validator_g_node_instance_id = predicate_validator(
        "GNodeInstanceId", property_format.is_uuid_canonical_textual
    )

    @validator("Strategy")
    def _validator_strategy(cls, v: StrategyName) -> StrategyName:
        return as_enum(v, StrategyName, StrategyName.NoActor)

    @validator("Status")
    def _validator_status(cls, v: GniStatus) -> GniStatus:
        return as_enum(v, GniStatus, GniStatus.Unknown)

    _validator_supervisor_container_id = predicate_validator(
        "SupervisorContainerId", property_format.is_uuid_canonical_textual
    )

    _validator_start_time_unix_s = predicate_validator(
        "StartTimeUnixS", property_format.is_reasonable_unix_time_s
    )

    @validator("AlgoAddress")
    def _validator_algo_address(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        if not property_format.is_algo_address_string_format(v):
            raise ValueError(f"AlgoAddress {v} must have AlgoAddressStringFormat")
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        d["GNode"] = self.GNode.as_dict()
        del d["Strategy"]
        Strategy = as_enum(self.Strategy, StrategyName, StrategyName.default())
        d["StrategyGtEnumSymbol"] = StrategyNameMap.local_to_type(Strategy)
        del d["Status"]
        Status = as_enum(self.Status, GniStatus, GniStatus.default())
        d["StatusGtEnumSymbol"] = GniStatusMap.local_to_type(Status)
        if d["AlgoAddress"] is None:
            del d["AlgoAddress"]
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class GNodeInstanceGt_Maker:
    type_name = "g.node.instance.gt"
    version = "000"

    def __init__(
        self,
        g_node_instance_id: str,
        g_node: GNodeGt,
        strategy: StrategyName,
        status: GniStatus,
        supervisor_container_id: str,
        start_time_unix_s: int,
        end_time_unix_s: int,
        algo_address: Optional[str],
    ):
        self.tuple = GNodeInstanceGt(
            GNodeInstanceId=g_node_instance_id,
            GNode=g_node,
            Strategy=strategy,
            Status=status,
            SupervisorContainerId=supervisor_container_id,
            StartTimeUnixS=start_time_unix_s,
            EndTimeUnixS=end_time_unix_s,
            AlgoAddress=algo_address,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: GNodeInstanceGt) -> str:
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> GNodeInstanceGt:
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> GNodeInstanceGt:
        d2 = dict(d)
        if "GNodeInstanceId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeInstanceId")
        if "GNode" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNode")
        if not isinstance(d2["GNode"], dict):
            raise SchemaError(f"d['GNode'] {d2['GNode']} must be a GNodeGt!")
        g_node = GNodeGt_Maker.dict_to_tuple(d2["GNode"])
        d2["GNode"] = g_node
        if "StrategyGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StrategyGtEnumSymbol")
        if d2["StrategyGtEnumSymbol"] in StrategyName000SchemaEnum.symbols:
            d2["Strategy"] = StrategyNameMap.type_to_local(d2["StrategyGtEnumSymbol"])
        else:
            d2["Strategy"] = StrategyName.default()
        if "StatusGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StatusGtEnumSymbol")
        if d2["StatusGtEnumSymbol"] in GniStatus000SchemaEnum.symbols:
            d2["Status"] = GniStatusMap.type_to_local(d2["StatusGtEnumSymbol"])
        else:
            d2["Status"] = GniStatus.default()
        if "SupervisorContainerId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SupervisorContainerId")
        if "StartTimeUnixS" not in d2.keys():
            raise SchemaError(f"dict {d2} missing StartTimeUnixS")
        if "EndTimeUnixS" not in d2.keys():
            raise SchemaError(f"dict {d2} missing EndTimeUnixS")
        if "AlgoAddress" not in d2.keys():
            d2["AlgoAddress"] = None
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return GNodeInstanceGt(
            GNodeInstanceId=d2["GNodeInstanceId"],
            GNode=d2["GNode"],
            Strategy=d2["Strategy"],
            Status=d2["Status"],
            SupervisorContainerId=d2["SupervisorContainerId"],
            StartTimeUnixS=d2["StartTimeUnixS"],
            EndTimeUnixS=d2["EndTimeUnixS"],
            AlgoAddress=d2["AlgoAddress"],
            TypeName=d2["TypeName"],
            Version="000",
        )

    @classmethod
    def tuple_to_dc(cls, t: GNodeInstanceGt) -> GNodeInstance:
        if t.GNodeInstanceId in GNodeInstance.by_id.keys():
            dc = GNodeInstance.by_id[t.GNodeInstanceId]
        else:
            dc = GNodeInstance(
                g_node_instance_id=t.GNodeInstanceId,
                g_node=t.GNode,
                strategy=t.Strategy,
                status=t.Status,
                supervisor_container_id=t.SupervisorContainerId,
                start_time_unix_s=t.StartTimeUnixS,
                end_time_unix_s=t.EndTimeUnixS,
                algo_address=t.AlgoAddress,
            )

        return dc

    @classmethod
    def dc_to_tuple(cls, dc: GNodeInstance) -> GNodeInstanceGt:
        t = GNodeInstanceGt_Maker(
            g_node_instance_id=dc.g_node_instance_id,
            g_node=dc.g_node,
            strategy=dc.strategy,
            status=dc.status,
            supervisor_container_id=dc.supervisor_container_id,
            start_time_unix_s=dc.start_time_unix_s,
            end_time_unix_s=dc.end_time_unix_s,
            algo_address=dc.algo_address,
        ).tuple
        return t

    @classmethod
    def type_to_dc(cls, t: str) -> GNodeInstance:
        return cls.tuple_to_dc(cls.type_to_tuple(t))

    @classmethod
    def dc_to_type(cls, dc: GNodeInstance) -> str:
        return cls.dc_to_tuple(dc).as_type()

    @classmethod
    def dict_to_dc(cls, d: dict[Any, str]) -> GNodeInstance:
        return cls.tuple_to_dc(cls.dict_to_tuple(d))



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MessageCategorySymbol(StrEnum):
    unknown = auto()
    rj = auto()
    rjb = auto()
    s = auto()
    gw = auto()
    post = auto()
    postack = auto()
    get = auto()

    @classmethod
    def default(cls) -> "MessageCategorySymbol":
        return cls.unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MessageCategory(StrEnum):
    Unknown = auto()
    RabbitJsonDirect = auto()
    RabbitJsonBroadcast = auto()
    RabbitGwSerial = auto()
    MqttJsonBroadcast = auto()
    RestApiPost = auto()
    RestApiPostResponse = auto()
    RestApiGet = auto()

    @classmethod
    def default(cls) -> "MessageCategory":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MarketPriceUnit(StrEnum):
    USDPerMWh = auto()

    @classmethod
    def default(cls) -> "MarketPriceUnit":
        return cls.USDPerMWh

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



""" AtnActorBase """
import functools
import logging
import time
from abc import abstractmethod
from typing import no_type_check

import pendulum
from gridworks.message import as_enum

from gwatn.config import Settings
from gwatn.enums import GNodeRole
from gwatn.enums import MessageCategorySymbol
from gwatn.enums import UniverseType
from gwatn.schemata import HeartbeatA
from gwatn.schemata import HeartbeatA_Maker
from gwatn.schemata import LatestPrice
from gwatn.schemata import LatestPrice_Maker
from gwatn.schemata import SimTimestep
from gwatn.schemata import SimTimestep_Maker
from gwatn.two_channel_actor_base import TwoChannelActorBase


LOG_FORMAT = (
    "%(levelname) -10s %(sasctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)

LOGGER.setLevel(logging.INFO)


class AtnActorBase(TwoChannelActorBase):
    def __init__(self, settings: Settings):
        super().__init__(settings=settings)
        self.settings: Settings = settings
        self.universe_type = as_enum(
            self.settings.universe_type_value, UniverseType, UniverseType.default()
        )
        self._time: float = self.get_initial_time_s()

    def local_rabbit_startup(self) -> None:
        rjb = MessageCategorySymbol.rjb.value
        tc_alias_lrh = self.settings.time_coordinator_alias.replace(".", "-")
        binding = f"{rjb}.{tc_alias_lrh}.timecoordinator.sim-timestep"

        cb = functools.partial(self.on_timecoordinator_bindok, binding=binding)
        self._consume_channel.queue_bind(
            self.queue_name, "timecoordinatormic_tx", routing_key=binding, callback=cb
        )
        LOGGER.info(
            f"Queue {self.queue_name} bound to timecoordinatormic_tx with {binding} "
        )
        self.strategy_rabbit_startup()

    def strategy_rabbit_startup(self) -> None:
        pass

    @no_type_check
    def on_timecoordinator_bindok(self, _unused_frame, binding) -> None:
        LOGGER.info(f"Queue {self.queue_name} bound with {binding}")

    def time(self) -> float:
        if self.universe_type == UniverseType.Dev:
            return self._time
        else:
            return time.time()

    def get_initial_time_s(self) -> float:
        if self.universe_type == UniverseType.Dev:
            return self.settings.initial_time_unix_s
        else:
            return time.time()

    def prepare_for_death(self) -> None:
        self.actor_main_stopped = True

    ########################
    ## Receives
    ########################

    def route_message(
        self, from_alias: str, from_role: GNodeRole, payload: HeartbeatA
    ) -> None:
        if payload.TypeName == SimTimestep_Maker.type_name:
            try:
                self.timestep_from_timecoordinator(payload)
            except:
                LOGGER.exception("Error in timestep_from_timecoordinator")
        elif payload.TypeName == LatestPrice_Maker.type_name:
            if from_role == GNodeRole.MarketMaker:
                try:
                    self.latest_price_from_market_maker(payload)
                except:
                    LOGGER.exception("Error in latest_price_from_market_maker")
        elif payload.TypeName == HeartbeatA_Maker.type_name:
            try:
                self.heartbeat_received(from_alias, payload)
            except:
                LOGGER.exception("Error in heartbeat_received")

    @abstractmethod
    def heartbeat_received(self, from_alias: str, payload: HeartbeatA) -> None:
        raise NotImplementedError

    @abstractmethod
    def latest_price_from_market_maker(self, payload: LatestPrice) -> None:
        raise NotImplementedError

    def timestep_from_timecoordinator(self, payload: SimTimestep):
        if self._time == 0:
            self._time = payload.TimeUnixS
            self.new_timestep(payload)
            LOGGER.info(f"TIME STARTED: {self.time_str()}")
        elif self._time < payload.TimeUnixS:
            self._time = payload.TimeUnixS
            self.new_timestep(payload)
            LOGGER.debug(f"Time is now {self.time_str()}")
        elif self._time == payload.TimeUnixS:
            self.repeat_timestep(payload)

    @abstractmethod
    def new_timestep(self, payload: SimTimestep) -> None:
        # LOGGER.info("New timestep in atn_actor_base")
        raise NotImplementedError

    @abstractmethod
    def repeat_timestep(self, payload: SimTimestep) -> None:
        # LOGGER.info("Timestep received again in atn_actor_base")
        raise NotImplementedError

    def time_str(self) -> str:
        return pendulum.from_timestamp(self.time()).strftime("%m/%d/%Y, %H:%M")



"""Type super.starter, version 000"""
import json
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import validator

from gwatn import property_format
from gwatn.schemata.g_node_instance_gt import GNodeInstanceGt
from gwatn.schemata.g_node_instance_gt import GNodeInstanceGt_Maker
from gwatn.schemata.supervisor_container_gt import SupervisorContainerGt
from gwatn.schemata.supervisor_container_gt import SupervisorContainerGt_Maker


class SuperStarter(BaseModel):
    SupervisorContainer: SupervisorContainerGt  #
    GniList: List[GNodeInstanceGt]
    #
    AliasWithKeyList: List[str]  #
    KeyList: List[str]  #
    TypeName: Literal["super.starter"] = "super.starter"
    Version: str = "000"

    @validator("GniList")
    def _validator_gni_list(cls, v: List) -> List:
        for elt in v:
            if not isinstance(elt, GNodeInstanceGt):
                raise ValueError(
                    f"elt {elt} of GniList must have type GNodeInstanceGt."
                )
        return v

    @validator("AliasWithKeyList")
    def _validator_alias_with_key_list(cls, v: List) -> List:
        for elt in v:
            if not property_format.is_lrd_alias_format(elt):
                raise ValueError(
                    f"failure of predicate is_lrd_alias_format() on elt {elt} of AliasWithKeyList"
                )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        d["SupervisorContainer"] = self.SupervisorContainer.as_dict()

        # Recursively call as_dict() for the SubTypes
        gni_list = []
        for elt in self.GniList:
            gni_list.append(elt.as_dict())
        d["GniList"] = gni_list
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class SuperStarter_Maker:
    type_name = "super.starter"
    version = "000"

    def __init__(
        self,
        supervisor_container: SupervisorContainerGt,
        gni_list: List[GNodeInstanceGt],
        alias_with_key_list: List[str],
        key_list: List[str],
    ):
        self.tuple = SuperStarter(
            SupervisorContainer=supervisor_container,
            GniList=gni_list,
            AliasWithKeyList=alias_with_key_list,
            KeyList=key_list,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: SuperStarter) -> str:
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> SuperStarter:
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> SuperStarter:
        d2 = dict(d)
        if "SupervisorContainer" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SupervisorContainer")
        if not isinstance(d2["SupervisorContainer"], dict):
            raise SchemaError(
                f"d['SupervisorContainer'] {d2['SupervisorContainer']} must be a SupervisorContainerGt!"
            )
        supervisor_container = SupervisorContainerGt_Maker.dict_to_tuple(
            d2["SupervisorContainer"]
        )
        d2["SupervisorContainer"] = supervisor_container
        if "GniList" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GniList")
        gni_list = []
        if not isinstance(d2["GniList"], List):
            raise SchemaError("GniList must be a List!")
        for elt in d2["GniList"]:
            if not isinstance(elt, dict):
                raise SchemaError(
                    f"elt {elt} of GniList must be "
                    "GNodeInstanceGt but not even a dict!"
                )
            gni_list.append(GNodeInstanceGt_Maker.dict_to_tuple(elt))
        d2["GniList"] = gni_list
        if "AliasWithKeyList" not in d2.keys():
            raise SchemaError(f"dict {d2} missing AliasWithKeyList")
        if "KeyList" not in d2.keys():
            raise SchemaError(f"dict {d2} missing KeyList")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return SuperStarter(
            SupervisorContainer=d2["SupervisorContainer"],
            GniList=d2["GniList"],
            AliasWithKeyList=d2["AliasWithKeyList"],
            KeyList=d2["KeyList"],
            TypeName=d2["TypeName"],
            Version="000",
        )



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class RecognizedCurrencyUnit(StrEnum):
    UNKNOWN = auto()
    USD = auto()
    GBP = auto()

    @classmethod
    def default(cls) -> "RecognizedCurrencyUnit":
        return cls.UNKNOWN

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class UniverseType(StrEnum):
    Dev = auto()
    Hybrid = auto()

    @classmethod
    def default(cls) -> "UniverseType":
        return cls.Dev

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class MixingValveFeedbackModel(StrEnum):
    ConstantSwt = auto()
    NaiveVariableSwt = auto()
    CautiousVariableSwt = auto()

    @classmethod
    def default(cls) -> "MixingValveFeedbackModel":
        return cls.ConstantSwt

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



import enum
import functools
import logging
import threading
import time
import uuid
from abc import ABC
from abc import abstractmethod
from enum import auto
from typing import Dict
from typing import Optional
from typing import no_type_check

import gridworks.property_format as property_format
import gridworks.utils as utils
import pika  # type: ignore
from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError

import gwatn.api_types as api_types
from gwatn.config import Settings
from gwatn.enums import GNodeRole
from gwatn.enums import MessageCategory
from gwatn.enums import MessageCategorySymbol
from gwatn.enums import UniverseType
from gwatn.schemata import HeartbeatA


class RabbitRole(StrEnum):
    atomictnode = auto()
    marketmaker = auto()
    supervisor = auto()
    timecoordinator = auto()
    world = auto()


RoleByRabbitRole: Dict[RabbitRole, GNodeRole] = {
    RabbitRole.atomictnode: GNodeRole.AtomicTNode,
    RabbitRole.marketmaker: GNodeRole.MarketMaker,
    RabbitRole.supervisor: GNodeRole.Supervisor,
    RabbitRole.timecoordinator: GNodeRole.TimeCoordinator,
    RabbitRole.world: GNodeRole.World,
}


RabbitRolebyRole: Dict[GNodeRole, RabbitRole] = {
    GNodeRole.AtomicTNode: RabbitRole.atomictnode,
    GNodeRole.MarketMaker: RabbitRole.marketmaker,
    GNodeRole.Supervisor: RabbitRole.supervisor,
    GNodeRole.TimeCoordinator: RabbitRole.timecoordinator,
    GNodeRole.World: RabbitRole.world,
}


class OnSendMessageDiagnostic(enum.Enum):
    CHANNEL_NOT_OPEN = "ChannelNotOpen"
    STOPPED_SO_NOT_SENDING = "StoppedSoNotSending"
    STOPPING_SO_NOT_SENDING = "StoppingSoNotSending"
    MESSAGE_SENT = "MessageSent"
    UNKNOWN_ERROR = "UnknownError"


class OnReceiveMessageDiagnostic(enum.Enum):
    TYPE_NAME_DECODING_PROBLEM = "TypeNameDecodingProblem"
    UNKNOWN_ROUTING_KEY_TYPE = "UnknownRoutingKeyType"
    UNHANDLED_ROUTING_KEY_TYPE = "UnhandledRoutingKeyType"
    UNKNOWN_MESSAGE_CATEGORY_SYMBOL = "UnknownMessageCategorySymbol"
    UNKNOWN_TYPE_NAME = "UnknownTypeName"
    FROM_GNODE_DECODING_PROBLEM = "FromGNodeDecodingProblem"
    UNKONWN_GNODE = "UnknownGNode"
    TO_DIRECT_ROUTING = "ToDirectRouting"


BACKOFF_NUMBER = 16

LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)

LOGGER.setLevel(logging.WARNING)


class TwoChannelActorBase(ABC):
    SHUTDOWN_INTERVAL = 0.1

    def __init__(
        self,
        settings: Settings,
    ):
        self.latest_routing_key: Optional[str] = None
        self.shutting_down: bool = False
        self.alias: str = settings.g_node_alias
        self.g_node_instance_id: str = settings.g_node_instance_id
        self.g_node_role: GNodeRole = GNodeRole(settings.g_node_role_value)
        self.rabbit_role: RabbitRole = RabbitRolebyRole[self.g_node_role]
        self.universe_type: UniverseType = UniverseType(settings.universe_type_value)
        self.actor_main_stopped: bool = False

        adder = "-F" + str(uuid.uuid4()).split("-")[0][0:3]
        self.queue_name: str = self.alias + adder
        self._consume_exchange: str = self.rabbit_role.value + "_tx"
        self._publish_exchange: str = self.rabbit_role.value + "mic_tx"

        self._consume_connection: Optional[
            pika.adapters.select_connection.SelectConnection
        ] = None
        self._consume_channel: Optional[pika.channel.Channel] = None
        self._closing_consumer: bool = False
        self._consumer_tag: Optional[str] = None
        self.should_reconnect_consumer: bool = False
        self.was_consuming: bool = False
        self._consuming: bool = False
        # In production, experiment with higher prefetch values
        # for higher consumer throughput
        self._prefetch_count: int = 1
        self._reconnect_delay: int = 0
        self._url: str = settings.rabbit.url.get_secret_value()

        self.is_debug_mode: bool = False
        self.consuming_thread: threading.Thread = threading.Thread(
            target=self.run_reconnecting_consumer
        )
        self.publishing_thread: threading.Thread = threading.Thread(
            target=self.run_publisher
        )
        self._publish_connection: Optional[
            pika.adapters.select_connection.SelectConnection
        ] = None
        self._publish_channel: Optional[pika.channel.Channel] = None
        self._stopping: bool = False
        self._stopped: bool = True
        self._latest_on_message_diagnostic: Optional[OnReceiveMessageDiagnostic] = None

    def start(self) -> None:
        self.consuming_thread.start()
        self.publishing_thread.start()
        self.local_start()
        self._stopped = False

    def local_start(self) -> None:
        """This should be overwritten in derived class for additional threads.
        It cannot assume the rabbit channels are established and that
        messages can be received or sent."""
        pass

    def stop(self) -> None:
        self.shutting_down = True
        self.prepare_for_death()
        while self.actor_main_stopped is False:
            time.sleep(self.SHUTDOWN_INTERVAL)
        self.stop_publisher()
        self.stop_consumer()
        self.local_stop()
        self.consuming_thread.join()
        self.publishing_thread.join()
        self._stopping = False
        self._stopped = True

    def local_stop() -> None:
        """This should be overwritten in derived class if there is a requirement
        to stop the additional threads started in local_start"""
        pass

    @no_type_check
    def on_message(self, _unused_channel, basic_deliver, properties, body) -> None:
        """Invoked by pika when a message is delivered from RabbitMQ. If a message
        does not get here that you expect should get here, check the routing key
        of the outbound message and the rabbitmq bindings.

        Parses the TypeName of the message payload and the GNodeAlias of the sender.
        If it recognizes the GNode and the TypeName then it sends the message on to
        the check_routing function, which will be defined in a child class (e.g., the
        GNodeFactoryActorBase if the actor is a GNodeFactory).

        From RabbitMQ:  The
        channel is passed for your convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.
        :param pika.channel.Channel _unused_channel: The channel object
        :param pika.Spec.Basic.Deliver: basic_deliver method
        :param pika.Spec.BasicProperties: properties
        :param bytes body: The message body
        """
        self.latest_routing_key = basic_deliver.routing_key
        LOGGER.debug(
            f"{self.alias}: Got {basic_deliver.routing_key} with delivery tag {basic_deliver.delivery_tag}"
        )
        self.acknowledge_message(basic_deliver.delivery_tag)
        try:
            type_name = self.get_payload_type_name(basic_deliver)
        except SchemaError:
            return
        self.type_name = type_name

        if type_name not in api_types.version_by_type_name().keys():
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.UNKNOWN_TYPE_NAME
            )
            LOGGER.warning(
                f"IGNORING MESSAGE. {self._latest_on_message_diagnostic}: {type_name}"
            )
            return

        try:
            payload = api_types.TypeMakerByName[type_name].type_to_tuple(body)
        except Exception as e:
            LOGGER.warning(
                f"TypeName for incoming message claimed to be {type_name}, but was not true! Failed to make a {api_types.TypeMakerByName[type_name].tuple}"
            )
            return

        routing_key: str = basic_deliver.routing_key

        try:
            from_alias = self.from_alias_from_routing_key(routing_key)
        except SchemaError as e:
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.FROM_GNODE_DECODING_PROBLEM
            )
            LOGGER.warning(
                f"IGNORING MESSAGE. {self._latest_on_message_diagnostic}: {e}"
            )
            return
        try:
            from_role = self.from_role_from_routing_key(routing_key)
        except SchemaError as e:
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.FROM_GNODE_DECODING_PROBLEM
            )
            LOGGER.warning(
                f"IGNORING MESSAGE. {self._latest_on_message_diagnostic}: {e}"
            )
            return

        self._latest_on_message_diagnostic = (
            OnReceiveMessageDiagnostic.TO_DIRECT_ROUTING
        )
        self.route_message(
            from_alias=from_alias,
            from_role=from_role,
            payload=payload,
        )

    @abstractmethod
    def route_message(
        self, from_alias: str, from_role: GNodeRole, payload: HeartbeatA
    ) -> None:
        raise NotImplementedError

    @no_type_check
    def send_message(
        self,
        payload: HeartbeatA,
        message_category: MessageCategory = MessageCategory.RabbitJsonDirect,
        to_role: Optional[GNodeRole] = None,
        to_g_node_alias: Optional[str] = None,
        radio_channel: Optional[str] = None,
    ) -> OnSendMessageDiagnostic:
        """Publish a direct message to another GNode in the registry world. The only type
        of direct messages in the registry use json (i.e. no more streamlined serial encoding),
        unlike in non-registry worlds.

        Args:
            payload: Any GridWorks schemata with a json content-type
            that includes TypeName as a json key, and has as_type()
            as an encoding method.
            routing_key_type: for creating routing key
            to_role (Optional[GNodeRole]): used if a direct message
            to_g_node_alias (str): used if a direct message

        Returns:
            OnSendMessageDiagnostic: MESSAGE_SENT with success, otherwise some
            description of why the message was not sent.
        """

        if self._stopping:
            return OnSendMessageDiagnostic.STOPPING_SO_NOT_SENDING
        if self._stopped:
            return OnSendMessageDiagnostic.STOPPED_SO_NOT_SENDING

        if "MessageId" in payload.as_dict():
            correlation_id = payload.MessageId
        else:
            correlation_id = str(uuid.uuid4())

        if message_category is MessageCategory.RabbitJsonDirect:
            if not isinstance(to_role, GNodeRole):
                raise Exception("Must include to_role for a direct message")
            if not property_format.is_lrd_alias_format(to_g_node_alias):
                raise Exception(
                    f"to_g_node_alias must have LrdAliasFormat. Got {to_g_node_alias}"
                )
            routing_key = self.direct_routing_key(
                to_role=to_role,
                payload=payload,
                to_g_node_alias=to_g_node_alias,
            )

            properties = pika.BasicProperties(
                reply_to=self.queue_name,
                app_id=self.alias,
                type=message_category.RabbitJsonDirect,
                correlation_id=correlation_id,
            )
        elif message_category is MessageCategory.RabbitJsonBroadcast:
            routing_key = self.broadcast_routing_key(
                payload=payload, radio_channel=radio_channel
            )
            properties = pika.BasicProperties(
                reply_to=self.queue_name,
                app_id=self.alias,
                type=message_category.RabbitJsonBroadcast,
                correlation_id=correlation_id,
            )
        else:
            raise Exception(f"Does not handle MessageCategory {message_category}")

        if self._publish_channel is None:
            LOGGER.error(f"No publish channel so not sending {routing_key}")
            return OnSendMessageDiagnostic.CHANNEL_NOT_OPEN
        if not self._publish_channel.is_open:
            LOGGER.error(f"Publish channel not open so not sending {routing_key}")
            return OnSendMessageDiagnostic.CHANNEL_NOT_OPEN

        try:
            self._publish_channel.basic_publish(
                exchange=self._publish_exchange,
                routing_key=routing_key,
                body=payload.as_type(),
                properties=properties,
            )
            LOGGER.debug(f" [x] Sent {payload.TypeName} w routing key {routing_key}")
            return OnSendMessageDiagnostic.MESSAGE_SENT

        except BaseException:
            LOGGER.exception("Problem w publish channel")
            return OnSendMessageDiagnostic.UNKNOWN_ERROR

    #####################
    # Abstract methods
    #####################

    @abstractmethod
    def prepare_for_death(self) -> None:
        """Once the agent is ready for its comms to be shut down it sets
        actor_main_stopped  to True. Write stoic code, with your
        agents ready for death at all times.  However, if there are threads running
        beyond the two designed for publishing and consuming messages, shut those
        down in this method."""
        raise NotImplementedError

    ########################
    # Core Rabbit infrastructure
    ########################

    def flush_consumer(self) -> None:
        self.should_reconnect_consumer = False
        self.was_consuming = False
        self._consume_connection = None
        self._consume_channel = None
        self._closing_consumer = False
        self._consumer_tag = None
        self._consuming = False

    def run_reconnecting_consumer(self) -> None:
        while not self.actor_main_stopped:
            self.run_consumer()
            self._maybe_reconnect_consumer()

    def _maybe_reconnect_consumer(self) -> None:
        if self.should_reconnect_consumer:
            self.stop_consumer()
            reconnect_delay = self._get_reconnect_delay()
            if not self.actor_main_stopped:
                LOGGER.info("Reconnecting after %d seconds", reconnect_delay)
            time.sleep(reconnect_delay)
            self.flush_consumer()

    def _get_reconnect_delay(self) -> int:
        if self.was_consuming:
            self._reconnect_delay = 0
        else:
            self._reconnect_delay += 1
        if self._reconnect_delay > 30:
            self._reconnect_delay = 30
        return self._reconnect_delay

    def connect_consumer(self) -> pika.SelectConnection:
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_consumer_connection_open method
        will be invoked by pika.
        :rtype: pika.SelectConnection
        """
        LOGGER.info("Connecting to %s", self._url)
        return pika.SelectConnection(
            parameters=pika.URLParameters(self._url),
            on_open_callback=self.on_consumer_connection_open,
            on_open_error_callback=self.on_consumer_connection_open_error,
            on_close_callback=self.on_consumer_connection_closed,
        )

    def close_consumer_connection(self) -> None:
        self._consuming = False
        if self._consume_connection:
            if (
                not self._consume_connection.is_closing
                and not self._consume_connection.is_closed
            ):
                LOGGER.info("Closing consume connection")
                self._consume_connection.close()

    @no_type_check
    def on_consumer_connection_open(self, _unused_connection) -> None:
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.
        :param pika.SelectConnection _unused_connection: The connection
        """
        LOGGER.info("Connection opened")
        self.open_consume_channel()

    @no_type_check
    def on_consumer_connection_open_error(self, _unused_connection, err) -> None:
        """This method is called by pika if the connection to RabbitMQ
        can't be established.
        :param pika.SelectConnection _unused_connection: The connection
        :param Exception err: The error
        """
        LOGGER.error(f"Consumer connection open failed: {err}")
        self.reconnect_consumer()

    @no_type_check
    def on_consumer_connection_closed(self, _unused_connection, reason) -> None:
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.
        :param pika.connection.Connection connection: The closed connection obj
        :param Exception reason: exception representing reason for loss of
            connection.
        """
        self._consume_channel = None
        if self._closing_consumer:
            self._consume_connection.ioloop.stop()
        else:
            LOGGER.warning(f"Consumer connection closed, reconnect necessary: {reason}")
            self.reconnect_consumer()

    def reconnect_consumer(self) -> None:
        """Will be invoked if the connection can't be opened or is
        closed. Indicates that a reconnect is necessary then stops the
        ioloop.
        """
        self.should_reconnect_consumer = True
        self.stop_consumer()

    def open_consume_channel(self) -> None:
        """Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.
        """
        LOGGER.info("Creating a new channel")
        self._consume_connection.channel(on_open_callback=self.on_consumer_channel_open)  # type: ignore

    @no_type_check
    def on_consumer_channel_open(self, channel) -> None:
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.
        Since the channel is now open, we'll declare the exchange to use.
        :param pika.channel.Channel channel: The channel object
        """
        LOGGER.info("Channel opened")
        self._consume_channel = channel
        self.add_on_consume_channel_close_callback()
        self.setup_exchange()

    @no_type_check
    def add_on_consume_channel_close_callback(self) -> None:
        """This method tells pika to call the on_consumer_channel_closed method if
        RabbitMQ unexpectedly closes the channel.
        """
        LOGGER.info("Adding consumer channel close callback")
        self._consume_channel.add_on_close_callback(self.on_consumer_channel_closed)

    @no_type_check
    def on_consumer_channel_closed(self, channel, reason) -> None:
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.
        :param pika.channel.Channel: The closed channel
        :param Exception reason: why the channel was closed
        """
        LOGGER.warning("Consume channel %i was closed: %s", channel, reason)
        self.close_consumer_connection()

    @no_type_check
    def setup_exchange(self) -> None:
        """Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
        command. When it is complete, the on_exchange_declareok method will
        be invoked by pika.
        :param str|unicode exchange_name: The name of the exchange to declare
        """
        LOGGER.info("Declaring exchange: %s", self._consume_exchange)
        # Note: using functools.partial is not required, it is demonstrating
        # how arbitrary data can be passed to the callback when it is called
        cb = functools.partial(
            self.on_exchange_declareok, userdata=self._consume_exchange
        )
        self._consume_channel.exchange_declare(
            exchange=self._consume_exchange,
            exchange_type="topic",
            durable=True,
            internal=True,
            callback=cb,
        )

    @no_type_check
    def on_exchange_declareok(self, _unused_frame, userdata) -> None:
        """Invoked by pika when RabbitMQ has finished the Exchange.Declare RPC
        command.
        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame
        :param str|unicode userdata: Extra user data (exchange name)
        """
        LOGGER.info("Exchange declared: %s", userdata)
        self.setup_queue()

    @no_type_check
    def setup_queue(self) -> None:
        """Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
        command. When it is complete, the on_queue_declareok method will
        be invoked by pika.
        :param str|unicode queue_name: The name of the queue to declare.
        """
        LOGGER.info(f"Declaring queue {self.queue_name}")
        cb = functools.partial(self.on_queue_declareok)
        self._consume_channel.queue_declare(
            queue=self.queue_name, auto_delete=True, callback=cb
        )

    @no_type_check
    def on_queue_declareok(self, _unused_frame) -> None:
        """Method invoked by pika when the Queue.Declare RPC call made in
        setup_queue has completed. In this method we will bind the queue
        and exchange together with the routing key by issuing the Queue.Bind
        RPC command. When this command is complete, the on_bindok method will
        be invoked by pika.
        :param pika.frame.Method _unused_frame: The Queue.DeclareOk frame
        :param str|unicode userdata: Extra user data (queue name)
        """
        lrh_alias = self.alias.replace(".", "-")
        rj = MessageCategorySymbol.rj.value
        direct_message_to_me_binding = f"{rj}.*.*.*.*.{lrh_alias}"

        LOGGER.info(
            "Binding %s to %s with %s",
            self._consume_exchange,
            self.queue_name,
            direct_message_to_me_binding,
        )
        cb = functools.partial(
            self.on_direct_message_bindok, binding=direct_message_to_me_binding
        )
        self._consume_channel.queue_bind(
            self.queue_name,
            self._consume_exchange,
            routing_key=direct_message_to_me_binding,
            callback=cb,
        )

    @no_type_check
    def on_direct_message_bindok(self, _unused_frame, binding) -> None:
        """Invoked by pika when the Queue.Bind method has completed for direct messages. At this
        point we will set the prefetch count for the channel.
        :param pika.frame.Method _unused_frame: The Queue.BindOk response frame
        :param str|unicode userdata: Extra user data (queue name)
        """
        LOGGER.info(f"Queue {self.queue_name} bound with {binding}")
        self.set_qos()

    @no_type_check
    def set_qos(self) -> None:
        """This method sets up the consumer prefetch to only be delivered
        one message at a time. The consumer must acknowledge this message
        before RabbitMQ will deliver another one. You should experiment
        with different prefetch values to achieve desired performance.
        """
        self._consume_channel.basic_qos(
            prefetch_count=self._prefetch_count, callback=self.on_basic_qos_ok
        )

    @no_type_check
    def on_basic_qos_ok(self, _unused_frame) -> None:
        """Invoked by pika when the Basic.QoS method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.
        :param pika.frame.Method _unused_frame: The Basic.QosOk response frame
        """
        LOGGER.info("QOS set to: %d", self._prefetch_count)
        self.local_rabbit_startup()
        self.start_consuming()

    def local_rabbit_startup(self) -> None:
        """This should be overwritten in derived class for any additional rabbit
        bindings. DO NOT start queues here"""
        pass

    @no_type_check
    def start_consuming(self) -> None:
        """This method sets up the consumer by first calling
        add_on_cancel_consumer_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming. The on_message method is passed in as a callback pika
        will invoke when a message is fully received.
        """
        LOGGER.info("Start consuming")
        self.add_on_cancel_consumer_callback()
        self._consumer_tag = self._consume_channel.basic_consume(
            self.queue_name, self.on_message
        )
        self.was_consuming = True
        self._consuming = True

    def add_on_cancel_consumer_callback(self) -> None:
        """Add a callback that will be invoked if RabbitMQ cancels the consumer
        for some reason. If RabbitMQ does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.
        """
        LOGGER.info("Adding consumer cancellation callback")
        self._consume_channel.add_on_cancel_callback(self.on_consumer_cancelled)  # type: ignore

    @no_type_check
    def on_consumer_cancelled(self, method_frame) -> None:
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.
        :param pika.frame.Method method_frame: The Basic.Cancel frame
        """
        LOGGER.info("Consumer was cancelled remotely, shutting down: %r", method_frame)
        if self._consume_channel:
            self._consume_channel.close()

    @no_type_check
    def acknowledge_message(self, delivery_tag) -> None:
        """Acknowledge the message delivery from RabbitMQ by sending a
        Basic.Ack RPC method for the delivery tag.
        :param int delivery_tag: The delivery tag from the Basic.Deliver frame
        """
        LOGGER.debug(
            f"Acknowledging message {delivery_tag}",
        )
        self._consume_channel.basic_ack(delivery_tag)

    def stop_consuming(self) -> None:
        """Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.
        """
        if self._consume_channel:
            LOGGER.info("Sending a Basic.Cancel RPC command to RabbitMQ")
            cb = functools.partial(
                self.on_cancelconsumer_ok, userdata=self._consumer_tag
            )
            self._consume_channel.basic_cancel(self._consumer_tag, cb)

    @no_type_check
    def on_cancelconsumer_ok(self, _unused_frame, userdata) -> None:
        """This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_consumer_channel_closed method once the channel has been
        closed, which will in-turn close the connection.
        :param pika.frame.Method _unused_frame: The Basic.CancelOk frame
        :param str|unicode userdata: Extra user data (consumer tag)
        """
        self._consuming = False
        LOGGER.info(
            "RabbitMQ acknowledged the cancellation of the consumer: %s", userdata
        )
        self.close_consumer_channel()
        self._closing_consumer = False

    def close_consumer_channel(self) -> None:
        """Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.
        """
        if self._consume_channel:
            if (
                not self._consume_channel.is_closing
                and not self._consume_channel.is_closed
            ):
                self._consume_channel.close()

    def run_consumer(self) -> None:
        """Run the example consumer by connecting to RabbitMQ and then
        starting the IOLoop to block and allow the SelectConnection to operate.
        """

        self._consume_connection = self.connect_consumer()
        self._consume_connection.ioloop.start()

    def stop_consumer(self) -> None:
        """Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelconsumer_ok
        will be invoked by pika, which will then closing the channel and
        connection. If you want to use this with CTRL-C, figure out
        how to add back the commented out ioloop.start() below without error.
        """
        if not self._closing_consumer:
            self._closing_consumer = True
            LOGGER.info("Consumer connection stopping")
            if self._consuming:
                self.stop_consuming()
                # self._consume_connection.ioloop.start()
            else:
                self._consume_connection.ioloop.stop()  # type: ignore

            LOGGER.info("Consumer connection stopped")

    def connect_publisher(self) -> pika.SelectConnection:
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.
        :rtype: pika.SelectConnection
        """
        LOGGER.info("Setting up publisher connection to %s", self._url)
        return pika.SelectConnection(
            pika.URLParameters(self._url),
            on_open_callback=self.on_publish_connection_open,
            on_open_error_callback=self.on_publish_connection_open_error,
            on_close_callback=self.on_publish_connection_closed,
        )

    @no_type_check
    def on_publish_connection_open(self, _unused_connection) -> None:
        """This method is called by pika once the publisher connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.
        :param pika.SelectConnection _unused_connection: The connection
        """
        LOGGER.info("Producer connection opened")
        self.open_publish_channel()

    @no_type_check
    def on_publish_connection_open_error(self, _unused_connection, err) -> None:
        """This method is called by pika if the connection to RabbitMQ
        can't be established.
        :param pika.SelectConnection _unused_connection: The connection
        :param Exception err: The error
        """
        LOGGER.error("Producer connection open failed, reopening in 1 second: %s", err)
        self._publish_connection.ioloop.call_later(
            1, self._publish_connection.ioloop.stop
        )

    @no_type_check
    def on_publish_connection_closed(self, _unused_connection, reason) -> None:
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.
        :param pika.connection.Connection connection: The closed connection obj
        :param Exception reason: exception representing reason for loss of
            connection.
        """
        self._publish_channel = None
        if self._stopping:
            self._publish_connection.ioloop.stop()
        else:
            LOGGER.warning("Connection closed, reopening in 1 second: %s", reason)
            self._publish_connection.ioloop.call_later(
                1, self._publish_connection.ioloop.stop
            )

    def open_publish_channel(self) -> None:
        """This method will open a new channel with RabbitMQ by issuing the
        Channel.Open RPC command. When RabbitMQ confirms the channel is open
        by sending the Channel.OpenOK RPC reply, the on_channel_open method
        will be invoked.
        """
        LOGGER.info("Creating a new publish channel")
        self._publish_connection.channel(on_open_callback=self.on_publish_channel_open)  # type: ignore

    @no_type_check
    def on_publish_channel_open(self, channel) -> None:
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.
        Since the channel is now open, we'll declare the exchange to use.
        :param pika.channel.Channel channel: The channel object
        """
        LOGGER.info("Publish channel opened")
        self._publish_channel = channel
        self.add_on_publish_channel_close_callback()

    def add_on_publish_channel_close_callback(self) -> None:
        """This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.
        """
        LOGGER.info("Adding channel close callback")
        self._publish_channel.add_on_close_callback(self.on_publish_channel_closed)  # type: ignore

    @no_type_check
    def on_publish_channel_closed(self, channel, reason) -> None:
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.
        :param pika.channel.Channel channel: The closed channel
        :param Exception reason: why the channel was closed
        """
        LOGGER.warning(f"Publish channel {channel} was closed: {reason}")
        self._publish_channel = None
        if not self._stopping:
            self._publish_connection.close()

    def run_publisher(self) -> None:
        """Run the example code by connecting and then starting the IOLoop."""
        while not self._stopping:
            self._publish_connection = None
            try:
                self._publish_connection = self.connect_publisher()
                self._publish_connection.ioloop.start()
            except KeyboardInterrupt:
                self.stop_publisher()
                if (
                    self._publish_connection is not None
                    and not self._publish_connection.is_closed
                ):
                    # Finish closing
                    self._publish_connection.ioloop.start()

        LOGGER.info("Stopped")

    def stop_publisher(self) -> None:
        """Stop the example by closing the channel and connection. We
        set a flag here so that we stop scheduling new messages to be
        published. The IOLoop is started because this method is
        invoked by the Try/Catch below when KeyboardInterrupt is caught.
        Starting the IOLoop again will allow the publisher to cleanly
        disconnect from RabbitMQ.
        """
        LOGGER.info(
            "Stopping RabbitMq message production - closing channel and connection"
        )
        self._stopping = True
        self.close_publish_channel()
        self.close_publish_connection()

    def close_publish_channel(self) -> None:
        """Invoke this command to close the channel with RabbitMQ by sending
        the Channel.Close RPC command.
        """
        if self._publish_channel:
            if (
                not self._publish_channel.is_closing
                and not self._publish_channel.is_closed
            ):
                self._publish_channel.close()

    def close_publish_connection(self) -> None:
        """This method closes the production connection to RabbitMQ."""
        if self._publish_connection:
            if (
                not self._publish_connection.is_closing
                and not self._publish_connection.is_closed
            ):
                self._publish_connection.close()
                LOGGER.info("Closing publish connection")

    ########################
    # Message passing semantics
    ########################

    @no_type_check
    def get_payload_type_name(self, basic_deliver) -> str:
        """The TypeName tis a string hat provides the strongly typed specification
            (API/ABI) for the incoming message. This is similar to knowing
            the protobuf name/method or the ABI name/method.

            The TypeName will articulate, in particular, how
            to decode the payload.

        Args:
            basic_deliver: the rabbit basic_deliver object
            body: the rabbit body object (i.e. the payload as incoming type)

        Returns:
            str: raises SchemaError if the TypeName is not accessible.
            Otherwise returns the TypeName

        """
        # TODO: right now we encode this in the routing key. However, we
        # could also add it to a different basic_deliver property, which
        # might be easier for developers to grock.

        try:
            type_name = self.type_name_from_routing_key(
                routing_key=basic_deliver.routing_key,
            )
        except SchemaError as e:
            LOGGER.info(f"Could not figure out TypeName: {e}")
            raise SchemaError(f"{e}")
        return type_name

    def broadcast_routing_key(
        self, payload: HeartbeatA, radio_channel: Optional[str]
    ) -> str:
        msg_type = MessageCategorySymbol.rjb.value
        from_alias_lrh = self.alias.replace(".", "-")
        type_name_lrh = payload.TypeName.replace(".", "-")
        from_role = self.rabbit_role.value
        if radio_channel is None:
            return f"{msg_type}.{from_alias_lrh}.{from_role}.{type_name_lrh}"
        else:
            if not property_format.is_lrd_alias_format(radio_channel):
                raise Exception(
                    f"radio_channel must have LrdAliasFormat. Got {radio_channel}"
                )
            return f"{msg_type}.{from_alias_lrh}.{from_role}.{type_name_lrh}.{radio_channel}"

    def direct_routing_key(
        self, to_role: GNodeRole, payload: HeartbeatA, to_g_node_alias: str
    ) -> str:
        msg_type = MessageCategorySymbol.rj.value
        from_lrh_alias = self.alias.replace(".", "-")
        from_role = self.rabbit_role.value
        to_role_val = RabbitRolebyRole[to_role].value
        to_lrh_alias = to_g_node_alias.replace(".", "-")
        type_name_lrh = payload.TypeName.replace(".", "-")

        direct_routing_key = f"{msg_type}.{from_lrh_alias}.{from_role}.{type_name_lrh}.{to_role_val}.{to_lrh_alias}"
        return direct_routing_key

    def type_name_from_routing_key(
        self,
        routing_key: str,
    ) -> str:
        """Returns the type alias of the message given the routing key. Raises a SchemaError
        exception if there is a problem decoding the type alias, or if it does not have
        the appropriate left-right-dot format.

        For all the GNode Registry messages, this involves json dumping the payload
        and looking for the TypeName dict key. However, in GridWorks World registries
        where the participating entities represent TerminalAssets, AtomicTNodes, and
        other operational actors in the electric grid model, there are a couple edge patterns
        of routing key where the TypeName is a substring of the routing key itself.

        The decision to embed the routing key INSIDE the body for the main set of
        messages has to do with the length limit of 255 for rabbit routing keys, and
        the desire to include both the FromGNodeAlias and the ToGNodeAlias in that key.

        Args:
            routing_key (str): This is the basic_deliver.routing_key string
            in a rabbit message

        Returns:
            str: the TypeName of the payload, in Lrd format
        """
        routing_key_words = routing_key.split(".")
        if len(routing_key_words) < 4:
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.UNKNOWN_ROUTING_KEY_TYPE
            )
            raise SchemaError(f"Routing key {routing_key} must have at least 3 words!")
        msg_category_symbol_value = routing_key_words[0]
        type_name_lrh = routing_key_words[3]

        try:
            msg_category_symbol = MessageCategorySymbol(msg_category_symbol_value)
        except ValueError:
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.UNKNOWN_MESSAGE_CATEGORY_SYMBOL
            )
            raise SchemaError(
                f"First  word of {routing_key} not a known MessageCategorySymbol!"
            )
        msg_category = utils.message_category_from_symbol(msg_category_symbol)
        allowable_categories = [
            MessageCategory.RabbitJsonDirect,
            MessageCategory.RabbitJsonBroadcast,
            MessageCategory.MqttJsonBroadcast,
        ]
        if msg_category not in allowable_categories:
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.UNHANDLED_ROUTING_KEY_TYPE
            )
            raise SchemaError(
                f"Rabbit messages do not handle {msg_category_symbol.value}"
            )

        if not property_format.is_lrh_alias_format(type_name_lrh):
            self._latest_on_message_diagnostic = (
                OnReceiveMessageDiagnostic.TYPE_NAME_DECODING_PROBLEM
            )
            raise SchemaError(
                f"TypeName {type_name_lrh} in {routing_key} message not lrh_alias_format!"
            )
        type_name = type_name_lrh.replace("-", ".")
        return type_name

    def from_alias_from_routing_key(self, routing_key: str) -> str:
        """Returns the GNodeAlias in left-right-dot format. Raises a SchemaError
        if there is trouble getting this.
        Args:
            routing_key (str): This is the basic_deliver.routing_key string
            in a rabbit message
        """
        routing_key_words = routing_key.split(".")
        try:
            from_g_node_alias_lrh = routing_key_words[1]
        except:
            raise SchemaError(f"{routing_key} must have at least three words!")

        if not property_format.is_lrh_alias_format(from_g_node_alias_lrh):
            raise SchemaError(
                f"GNodeAlias not is_lrh_alias_format for routing key {routing_key}!"
            )
        from_g_node_alias = from_g_node_alias_lrh.replace("-", ".")
        return from_g_node_alias

    def from_role_from_routing_key(self, routing_key: str) -> GNodeRole:
        """Returns the GNodeRole that the message came from. Raises a SchemaError
        if there is trouble getting this.
        Args:
            routing_key (str): basic_deliver.routing_key string
            in a rabbit message

        Raises:
            SchemaError: Error in message construction

        Returns:
            GNodeRole: GNodeRole of the actor that sent the message
        """
        routing_key_words = routing_key.split(".")
        try:
            from_g_node_rabbit_role = routing_key_words[2]
        except:
            raise SchemaError(f"{routing_key} must have at least three words!")
        try:
            rabbit_role = RabbitRole(from_g_node_rabbit_role)
        except ValueError:
            raise SchemaError(
                f"Unknown short alias {rabbit_role} in {routing_key}"
                f" Must belong to {RabbitRole}"
            )

        return RoleByRabbitRole[rabbit_role]

    #################################################
    # On receiving messages broadcast to all listners
    #################################################

    #################################################
    # Various
    #################################################

    def __repr__(self) -> str:
        return f"{self.alias}"



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class SupervisorContainerStatus(StrEnum):
    Unknown = auto()
    Authorized = auto()
    Launching = auto()
    Provisioning = auto()
    Running = auto()
    Stopped = auto()
    Deleted = auto()

    @classmethod
    def default(cls) -> "SupervisorContainerStatus":
        return cls.Unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class RecognizedTemperatureUnit(StrEnum):
    C = auto()
    F = auto()

    @classmethod
    def default(cls) -> "RecognizedTemperatureUnit":
        return cls.C

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]



"""Type atn.params.report.heatpumpwithbooststore, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal
from typing import Optional

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import validator

from gwatn import property_format
from gwatn.property_format import predicate_validator
from gwatn.schemata.atn_params_heatpumpwithbooststore import (
    AtnParamsHeatpumpwithbooststore,
)
from gwatn.schemata.atn_params_heatpumpwithbooststore import (
    AtnParamsHeatpumpwithbooststore_Maker,
)


class AtnParamsReportHeatpumpwithbooststore(BaseModel):
    GNodeAlias: str  #
    GNodeInstanceId: str  #
    TimeUnixS: int  #
    IrlTimeUnixS: Optional[int] = None
    AtnParams: AtnParamsHeatpumpwithbooststore  #
    TypeName: Literal[
        "atn.params.report.heatpumpwithbooststore"
    ] = "atn.params.report.heatpumpwithbooststore"
    Version: str = "000"

    _validator_g_node_alias = predicate_validator(
        "GNodeAlias", property_format.is_lrd_alias_format
    )

    _validator_g_node_instance_id = predicate_validator(
        "GNodeInstanceId", property_format.is_uuid_canonical_textual
    )

    _validator_time_unix_s = predicate_validator(
        "TimeUnixS", property_format.is_reasonable_unix_time_s
    )

    @validator("IrlTimeUnixS")
    def _validator_irl_time_unix_s(cls, v: Optional[int]) -> Optional[int]:
        if v is None:
            return v
        if not property_format.is_reasonable_unix_time_s(v):
            raise ValueError(f"IrlTimeUnixS {v} must have ReasonableUnixTimeS")
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        if d["IrlTimeUnixS"] is None:
            del d["IrlTimeUnixS"]
        d["AtnParams"] = self.AtnParams.as_dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class AtnParamsReportHeatpumpwithbooststore_Maker:
    type_name = "atn.params.report.heatpumpwithbooststore"
    version = "000"

    def __init__(
        self,
        g_node_alias: str,
        g_node_instance_id: str,
        time_unix_s: int,
        irl_time_unix_s: Optional[int],
        atn_params: AtnParamsHeatpumpwithbooststore,
    ):
        self.tuple = AtnParamsReportHeatpumpwithbooststore(
            GNodeAlias=g_node_alias,
            GNodeInstanceId=g_node_instance_id,
            TimeUnixS=time_unix_s,
            IrlTimeUnixS=irl_time_unix_s,
            AtnParams=atn_params,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: AtnParamsReportHeatpumpwithbooststore) -> str:
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> AtnParamsReportHeatpumpwithbooststore:
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> AtnParamsReportHeatpumpwithbooststore:
        d2 = dict(d)
        if "GNodeAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeAlias")
        if "GNodeInstanceId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeInstanceId")
        if "TimeUnixS" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TimeUnixS")
        if "IrlTimeUnixS" not in d2.keys():
            d2["IrlTimeUnixS"] = None
        if "AtnParams" not in d2.keys():
            raise SchemaError(f"dict {d2} missing AtnParams")
        if not isinstance(d2["AtnParams"], dict):
            raise SchemaError(
                f"d['AtnParams'] {d2['AtnParams']} must be a AtnParamsHeatpumpwithbooststore!"
            )
        atn_params = AtnParamsHeatpumpwithbooststore_Maker.dict_to_tuple(
            d2["AtnParams"]
        )
        d2["AtnParams"] = atn_params
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return AtnParamsReportHeatpumpwithbooststore(
            GNodeAlias=d2["GNodeAlias"],
            GNodeInstanceId=d2["GNodeInstanceId"],
            TimeUnixS=d2["TimeUnixS"],
            IrlTimeUnixS=d2["IrlTimeUnixS"],
            AtnParams=d2["AtnParams"],
            TypeName=d2["TypeName"],
            Version="000",
        )



"""Gridworks AtomicTNode"""
import gwatn.api_types as api_types
import gwatn.config as config
import gwatn.enums as enums
from gwatn.atn_actor_base import AtnActorBase


__all__ = [
    "api_types",
    "config",
    "conversion_factors",
    "enums",
    "errors",
    "message",
    "property_format",
    "AtnActorBase",
]



import string
import struct
from datetime import datetime
from typing import Any
from typing import Callable

import algosdk  # type: ignore
import pendulum
import pydantic

from gwatn.data_classes import MarketType


def is_world_instance_alias_format(candidate: str) -> bool:
    try:
        words = candidate.split("__")
    except:
        return False
    try:
        int(words[1])
    except:
        return False
    try:
        alias_words = words[0].split(".")
    except:
        return False
    if len(alias_words) > 1:
        return False
    return True


def predicate_validator(
    field_name: str, predicate: Callable[[Any], bool], error_format: str = ""
) -> classmethod:  # type: ignore
    def _validator(v: Any) -> Any:
        if not predicate(v):
            if error_format:
                err_str = error_format.format(value=v)
            else:
                err_str = f"{field_name}: {predicate} fails for [{v}]"
            raise ValueError(err_str)
        return v

    return pydantic.validator(field_name, allow_reuse=True)(_validator)


def is_iso_format(candidate: str) -> bool:
    try:
        datetime.fromisoformat(candidate.replace("Z", "+00:00"))
    except:
        return False
    return True


def is_algo_address_string_format(candidate: str) -> bool:
    """The public key of a private/public Ed25519 key pairis transformed into an Algorand address,
    by adding a 4-byte checksum to the end of the public key and then encoding it
    in base32.

    Args:
        candidate (str): candidate

    Returns:
        bool: True if is_algo_address_string_format
    """

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(candidate))
    except Exception as e:
        return False
    return True


def is_algo_msg_pack_encoded(candidate: str) -> bool:
    try:
        algosdk.encoding.future_msgpack_decode(candidate)
    except Exception as e:
        return False
    return True


def is_64_bit_hex(candidate: str) -> bool:
    if len(candidate) != 8:
        return False
    if not all(c in string.hexdigits for c in candidate):
        return False
    return True


def check_is_64_bit_hex(candidate: str) -> None:
    if len(candidate) != 8:
        raise ValueError(f" {candidate} Must be length 8, not {len(candidate)}")
    if not all(c in string.hexdigits for c in candidate):
        raise ValueError("Must be hex digits")


def is_bit(candidate: int) -> bool:
    if candidate not in {0, 1}:
        return False
    return True


def check_is_bit(candidate: int) -> None:
    if candidate not in {0, 1}:
        raise ValueError(f"{candidate} must be either 0 or 1")


def is_lrd_alias_format(candidate: str) -> bool:
    """Lowercase AlphanumericStrings separated by dots (i.e. periods), with most
    significant word to the left.  I.e. `d1.ne` is the child of `d1`.
    Checking the format cannot verify the significance of words. All
    words must be alphanumeric. Most significant word must start with
    an alphabet charecter

    Args:
        candidate (str): candidate

    Returns:
        bool: True if is_lrod_alias_format
    """
    try:
        x = candidate.split(".")
    except:
        return False
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        return False
    for word in x:
        if not word.isalnum():
            return False
    if not candidate.islower():
        return False
    return True


def check_is_lrd_alias_format(candidate: str) -> None:
    """Lowercase AlphanumericStrings separated by dots (i.e. periods), with most
    significant word to the left.  I.e. `d1.ne` is the child of `d1`.
    Checking the format cannot verify the significance of words. All
    words must be alphanumeric. Most significant word must start with
    an alphabet charecter


    Raises:
        ValueError: if candidate is not of lrd format (e.g. d1.iso.me.apple)
    """
    try:
        x = candidate.split(".")
    except:
        raise ValueError("Failed to seperate into words with split'.'")
    first_word: str = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(
            f"Most significant word must start with alphabet char. Got '{first_word}'"
        )
    for word in x:
        w: str = str(word)
        if not w.isalnum():
            raise ValueError(f"words seperated by dots must be alphanumeric. Got '{w}'")
    if not candidate.islower():
        raise ValueError(f"alias must be lowercase. Got '{candidate}'")


def is_lru_alias_format(candidate: str) -> bool:
    """AlphanumericStrings separated by underscores, with most
    significant word to the left.  I.e. `d1.ne` is the child of `d1`.
    Checking the format cannot verify the significance of words. All
    words must be alphanumeric. Most significant word must start with
    an alphabet charecter"""
    try:
        x = candidate.split("_")
    except:
        return False
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        return False
    for word in x:
        if not word.isalnum():
            return False
    if not candidate.islower():
        return False
    return True


def is_lrh_alias_format(candidate: str) -> bool:
    """AlphanumericStrings separated by hyphens, with most
    significant word to the left.  I.e. `d1.ne` is the child of `d1`.
    Checking the format cannot verify the significance of words. All
    words must be alphanumeric. Most significant word must start with
    an alphabet charecter"""
    try:
        x = candidate.split("-")
    except:
        return False
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        return False
    for word in x:
        if not word.isalnum():
            return False
    if not candidate.islower():
        return False
    return True


def is_positive_integer(candidate: Any) -> bool:
    if not isinstance(candidate, int):
        return False
    if candidate <= 0:
        return False
    return True


def check_is_positive_integer(candidate: int) -> None:
    if not isinstance(candidate, int):
        raise ValueError("Must be an integer")
    if candidate <= 0:
        raise ValueError("Must be positive integer")


def is_reasonable_unix_time_ms(candidate: int) -> bool:
    if pendulum.parse("2000-01-01T00:00:00Z").int_timestamp * 1000 > candidate:
        return False
    if pendulum.parse("3000-01-01T00:00:00Z").int_timestamp * 1000 < candidate:
        return False
    return True


def check_is_reasonable_unix_time_ms(candidate: int) -> None:
    if pendulum.parse("2000-01-01T00:00:00Z").int_timestamp * 1000 > candidate:
        raise ValueError("ReasonableUnixTimeMs must be after 2000 AD")
    if pendulum.parse("3000-01-01T00:00:00Z").int_timestamp * 1000 < candidate:
        raise ValueError("ReasonableUnixTimeMs must be before 3000 AD")


def is_reasonable_unix_time_s(candidate: int) -> bool:
    if pendulum.parse("2000-01-01T00:00:00Z").int_timestamp > candidate:
        return False
    if pendulum.parse("3000-01-01T00:00:00Z").int_timestamp < candidate:
        return False
    return True


def check_is_reasonable_unix_time_s(candidate: int) -> None:
    if pendulum.parse("2000-01-01T00:00:00Z").int_timestamp > candidate:
        raise ValueError("ReasonableUnixTimeS must be after 2000 AD")
    if pendulum.parse("3000-01-01T00:00:00Z").int_timestamp < candidate:
        raise ValueError("ReasonableUnixTimeS must be before 3000 AD")


def is_unsigned_short(candidate: int) -> bool:
    try:
        struct.pack("H", candidate)
    except:
        return False
    return True


def check_is_unsigned_short(candidate: int) -> None:
    try:
        struct.pack("H", candidate)
    except:
        raise ValueError("requires 0 <= number <= 65535")


def is_short_integer(candidate: int) -> bool:
    try:
        struct.pack("h", candidate)
    except:
        return False
    return True


def check_is_short_integer(candidate: int) -> None:
    try:
        struct.pack("h", candidate)
    except:
        raise ValueError("short format requires (-32767 -1) <= number <= 32767")


def is_uuid_canonical_textual(candidate: str) -> bool:
    try:
        x = candidate.split("-")
    except AttributeError:
        return False
    if len(x) != 5:
        return False
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            return False
    if len(x[0]) != 8:
        return False
    if len(x[1]) != 4:
        return False
    if len(x[2]) != 4:
        return False
    if len(x[3]) != 4:
        return False
    if len(x[4]) != 12:
        return False
    return True


def check_is_uuid_canonical_textual(candidate: str) -> None:
    try:
        x = candidate.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"Did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError("Words are not all hex")
    if len(x[0]) != 8:
        raise ValueError("Word 0  not of length 8")
    if len(x[1]) != 4:
        raise ValueError("Word 1 not of length 4")
    if len(x[2]) != 4:
        raise ValueError("Word 2 not of length 4")
    if len(x[3]) != 4:
        raise ValueError("Word 3 not of length 4")
    if len(x[4]) != 12:
        raise ValueError("Word 4 not of length 12")


def is_market_type_alias_lrd_format(candidate: str) -> bool:
    try:
        x = candidate.split(".")
    except AttributeError:
        return False
    if not x[0] in MarketType.by_id.keys():
        return False
    g_node_alias = ".".join(x[1:])
    return is_lrd_alias_format(g_node_alias)


def is_market_slot_name_lrd_format(candidate: str) -> bool:
    try:
        x = candidate.split(".")
    except AttributeError:
        return False
    slot_start = x[-1]
    if len(slot_start) != 10:
        return False
    try:
        slot_start = int(slot_start)
    except ValueError:
        return False
    if slot_start % 300 != 0:
        return False

    market_type_alias_lrd = ".".join(x[:-1])
    if not is_market_type_alias_lrd_format(market_type_alias_lrd):
        return False

    market_type = MarketType.by_id[market_type_alias_lrd.split(".")[0]]
    if not slot_start % (market_type.duration_minutes * 60) == 0:
        print(
            f"market_slot_start_s mod {(market_type.duration_minutes * 60)} must be 0"
        )
        return False

    return True


def check_world_alias_matches_universe(g_node_alias: str, universe: str) -> None:
    """
    Raises:
        ValueError: if g_node_alias is not LRD format or if first word does not match universe
    """
    check_is_lrd_alias_format(g_node_alias)
    world_alias = g_node_alias.split(".")[0]
    if universe == "dev":
        if world_alias[0] != "d":
            raise ValueError(
                f"World alias for dev universe must start with d. Got {world_alias}"
            )



"""Type market.price, version 000"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import validator

from gwatn.enums import MarketPriceUnit


class MarketPriceUnit000SchemaEnum:
    enum_name: str = "market.price.unit.000"
    symbols: List[str] = [
        "00000000",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class MarketPriceUnit000(StrEnum):
    USDPerMWh = auto()

    @classmethod
    def default(cls) -> "MarketPriceUnit000":
        return cls.USDPerMWh

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class MarketPriceUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> MarketPriceUnit:
        if not MarketPriceUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to MarketPriceUnit000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, MarketPriceUnit, MarketPriceUnit.default())

    @classmethod
    def local_to_type(cls, market_price_unit: MarketPriceUnit) -> str:
        if not isinstance(market_price_unit, MarketPriceUnit):
            raise SchemaError(f"{market_price_unit} must be of type {MarketPriceUnit}")
        versioned_enum = as_enum(
            market_price_unit, MarketPriceUnit000, MarketPriceUnit000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, MarketPriceUnit000] = {
        "00000000": MarketPriceUnit000.USDPerMWh,
    }

    versioned_enum_to_type_dict: Dict[MarketPriceUnit000, str] = {
        MarketPriceUnit000.USDPerMWh: "00000000",
    }


class MarketPrice(BaseModel):
    ValueTimes1000: int  #
    Unit: MarketPriceUnit  #
    TypeName: Literal["market.price"] = "market.price"
    Version: str = "000"

    @validator("Unit")
    def _validator_unit(cls, v: MarketPriceUnit) -> MarketPriceUnit:
        return as_enum(v, MarketPriceUnit, MarketPriceUnit.USDPerMWh)

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        del d["Unit"]
        Unit = as_enum(self.Unit, MarketPriceUnit, MarketPriceUnit.default())
        d["UnitGtEnumSymbol"] = MarketPriceUnitMap.local_to_type(Unit)
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class MarketPrice_Maker:
    type_name = "market.price"
    version = "000"

    def __init__(self, value_times1000: int, unit: MarketPriceUnit):
        self.tuple = MarketPrice(
            ValueTimes1000=value_times1000,
            Unit=unit,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: MarketPrice) -> str:
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> MarketPrice:
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> MarketPrice:
        d2 = dict(d)
        if "ValueTimes1000" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ValueTimes1000")
        if "UnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing UnitGtEnumSymbol")
        if d2["UnitGtEnumSymbol"] in MarketPriceUnit000SchemaEnum.symbols:
            d2["Unit"] = MarketPriceUnitMap.type_to_local(d2["UnitGtEnumSymbol"])
        else:
            d2["Unit"] = MarketPriceUnit.default()
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return MarketPrice(
            ValueTimes1000=d2["ValueTimes1000"],
            Unit=d2["Unit"],
            TypeName=d2["TypeName"],
            Version="000",
        )



from enum import auto
from typing import List

from fastapi_utils.enums import StrEnum


class Role(StrEnum):
    PrimaryScada = auto()
    HomeAlone = auto()
    Heatpump = auto()
    MultipurposeSensor = auto()
    CurrentTransformer = auto()
    RadiatorFan = auto()
    PipeFlowMeter = auto()
    BaseboardRadiator = auto()
    CirculatorPump = auto()
    DedicatedThermalStore = auto()
    HeatedSpace = auto()
    HydronicPipe = auto()
    PrimaryMeter = auto()
    Outdoors = auto()
    BoostElement = auto()
    BooleanActuator = auto()
    Atn = auto()
    TankWaterTempSensor = auto()
    PipeTempSensor = auto()
    RoomTempSensor = auto()
    OutdoorTempSensor = auto()

    @classmethod
    def default(cls) -> "Role":
        return cls.HeatedSpace

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]
